from typing import Literal, Optional, Union

from typing_extensions import Annotated

from .._internal.common_nodes import Node
from .._internal.io import WithSuffix
from .._internal.io_basics import AbsoluteFilePath as AbsoluteFilePath
from .._internal.url import HttpUrl
from ..generic.v0_2 import VALID_COVER_IMAGE_EXTENSIONS as VALID_COVER_IMAGE_EXTENSIONS
from ..generic.v0_2 import AttachmentsDescr as AttachmentsDescr
from ..generic.v0_2 import Author as Author
from ..generic.v0_2 import BadgeDescr as BadgeDescr
from ..generic.v0_2 import CiteEntry as CiteEntry
from ..generic.v0_2 import Doi as Doi
from ..generic.v0_2 import GenericDescrBase
from ..generic.v0_2 import LinkedResource as LinkedResource
from ..generic.v0_2 import Maintainer as Maintainer
from ..generic.v0_2 import OrcidId as OrcidId
from ..generic.v0_2 import RelativeFilePath as RelativeFilePath
from ..generic.v0_2 import ResourceId as ResourceId
from ..generic.v0_2 import Uploader as Uploader
from ..generic.v0_2 import Version as Version


class NotebookId(ResourceId):
    pass


_WithNotebookSuffix = WithSuffix(".ipynb", case_sensitive=True)
NotebookSource = Union[
    Annotated[HttpUrl, _WithNotebookSuffix],
    Annotated[AbsoluteFilePath, _WithNotebookSuffix],
    Annotated[RelativeFilePath, _WithNotebookSuffix],
]


class NotebookDescr(GenericDescrBase, title="bioimage.io notebook specification"):
    """Bioimage.io description of a Jupyter Notebook."""

    type: Literal["notebook"] = "notebook"

    id: Optional[NotebookId] = None
    """Model zoo (bioimage.io) wide, unique identifier (assigned by bioimage.io)"""

    source: NotebookSource
    """The Jupyter notebook"""


class LinkedNotebook(Node):
    """Reference to a bioimage.io notebook."""

    id: NotebookId
    """A valid notebook `id` from the bioimage.io collection."""

    version_number: Optional[int] = None
    """version number (n-th published version, not the semantic version) of linked notebook"""



from argparse import ArgumentParser
from pathlib import Path
from typing import Literal

from typing_extensions import assert_never

from bioimageio.spec import load_description_and_validate_format_only

# from tests.test_bioimageio_collection import (
#     KNOWN_INVALID,
#     KNOWN_INVALID_AS_LATEST,
#     RDF_BASE_URL,
# )
RDF_BASE_URL = "https://bioimage-io.github.io/collection-bioimage-io/rdfs/"


KNOWN_INVALID = {
    "10.5281/zenodo.5749843/5888237/rdf.yaml",
    "10.5281/zenodo.5910163/5942853/rdf.yaml",
    "10.5281/zenodo.5910854/6539073/rdf.yaml",
    "10.5281/zenodo.5914248/6514622/rdf.yaml",
    "10.5281/zenodo.6559929/6559930/rdf.yaml",
    "10.5281/zenodo.7614645/7642674/rdf.yaml",
    "biapy/biapy/latest/rdf.yaml",
    "biapy/notebook_classification_2d/latest/rdf.yaml",
    "biapy/Notebook_semantic_segmentation_3d/latest/rdf.yaml",
    "deepimagej/deepimagej/latest/rdf.yaml",
    "deepimagej/DeepSTORMZeroCostDL4Mic/latest/rdf.yaml",
    "deepimagej/Mt3VirtualStaining/latest/rdf.yaml",
    "deepimagej/MU-Lux_CTC_PhC-C2DL-PSC/latest/rdf.yaml",
    "deepimagej/SkinLesionClassification/latest/rdf.yaml",
    "deepimagej/SMLMDensityMapEstimationDEFCoN/latest/rdf.yaml",
    "deepimagej/UNet2DGlioblastomaSegmentation/latest/rdf.yaml",
    "deepimagej/WidefieldDapiSuperResolution/latest/rdf.yaml",
    "deepimagej/WidefieldFitcSuperResolution/latest/rdf.yaml",
    "deepimagej/WidefieldTxredSuperResolution/latest/rdf.yaml",
    "fiji/N2VSEMDemo/latest/rdf.yaml",
    "ilastik/mitoem_segmentation_challenge/latest/rdf.yaml",
    "imjoy/LuCa-7color/latest/rdf.yaml",
    "zero/Dataset_CARE_2D_coli_DeepBacs/latest/rdf.yaml",
    "zero/Dataset_fnet_DeepBacs/latest/rdf.yaml",
    "zero/Dataset_Noise2Void_2D_subtilis_DeepBacs/latest/rdf.yaml",
    "zero/Dataset_SplineDist_2D_DeepBacs/latest/rdf.yaml",
    "zero/Dataset_StarDist_2D_DeepBacs/latest/rdf.yaml",
    "zero/Dataset_U-Net_2D_DeepBacs/latest/rdf.yaml",
    "zero/Dataset_U-Net_2D_multilabel_DeepBacs/latest/rdf.yaml",
    "zero/Dataset_YOLOv2_antibiotic_DeepBacs/latest/rdf.yaml",
    "zero/Dataset_YOLOv2_coli_DeepBacs/latest/rdf.yaml",
    "zero/Notebook_CycleGAN_2D_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Notebook_DecoNoising_2D_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Notebook_Detectron2_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Notebook_DRMIME_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Notebook_EmbedSeg_2D_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Notebook_MaskRCNN_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Notebook_pix2pix_2D_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Notebook_RetinaNet_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Notebook_StarDist_3D_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Notebook_U-Net_2D_multilabel_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Notebook_U-Net_2D_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Notebook_U-Net_3D_ZeroCostDL4Mic/latest/rdf.yaml",
}
KNOWN_INVALID_AS_LATEST = {
    "10.5281/zenodo.5749843/5888237/rdf.yaml",
    "10.5281/zenodo.5874841/6630266/rdf.yaml",
    "10.5281/zenodo.5910163/5942853/rdf.yaml",
    "10.5281/zenodo.5914248/6514622/rdf.yaml",
    "10.5281/zenodo.5914248/8186255/rdf.yaml",
    "10.5281/zenodo.6383429/7774505/rdf.yaml",
    "10.5281/zenodo.6406803/6406804/rdf.yaml",
    "10.5281/zenodo.6559474/6559475/rdf.yaml",
    "10.5281/zenodo.6559929/6559930/rdf.yaml",
    "10.5281/zenodo.6811491/6811492/rdf.yaml",
    "10.5281/zenodo.6865412/6919253/rdf.yaml",
    "10.5281/zenodo.7380171/7405349/rdf.yaml",
    "10.5281/zenodo.7614645/7642674/rdf.yaml",
    "10.5281/zenodo.8401064/8429203/rdf.yaml",
    "10.5281/zenodo.8421755/8432366/rdf.yaml",
    "biapy/biapy/latest/rdf.yaml",
    "biapy/notebook_classification_2d/latest/rdf.yaml",
    "biapy/notebook_classification_3d/latest/rdf.yaml",
    "biapy/notebook_denoising_2d/latest/rdf.yaml",
    "biapy/notebook_denoising_3d/latest/rdf.yaml",
    "biapy/notebook_detection_2d/latest/rdf.yaml",
    "biapy/notebook_detection_3d/latest/rdf.yaml",
    "biapy/notebook_instance_segmentation_2d/latest/rdf.yaml",
    "biapy/notebook_instance_segmentation_3d/latest/rdf.yaml",
    "biapy/notebook_self_supervision_2d/latest/rdf.yaml",
    "biapy/notebook_self_supervision_3d/latest/rdf.yaml",
    "biapy/notebook_semantic_segmentation_2d/latest/rdf.yaml",
    "biapy/Notebook_semantic_segmentation_3d/latest/rdf.yaml",
    "biapy/notebook_super_resolution_2d/latest/rdf.yaml",
    "biapy/notebook_super_resolution_3d/latest/rdf.yaml",
    "bioimageio/stardist/latest/rdf.yaml",
    "deepimagej/deepimagej-web/latest/rdf.yaml",
    "deepimagej/deepimagej/latest/rdf.yaml",
    "deepimagej/DeepSTORMZeroCostDL4Mic/latest/rdf.yaml",
    "deepimagej/DeepSTORMZeroCostDL4Mic/latest/rdf.yaml",
    "deepimagej/DeepSTORMZeroCostDL4Mic/latest/rdf.yaml",
    "deepimagej/DeepSTORMZeroCostDL4Mic/latest/rdf.yaml",
    "deepimagej/EVsTEMsegmentationFRUNet/latest/rdf.yaml",
    "deepimagej/MoNuSeg_digital_pathology_miccai2018/latest/rdf.yaml",
    "deepimagej/Mt3VirtualStaining/latest/rdf.yaml",
    "deepimagej/MU-Lux_CTC_PhC-C2DL-PSC/latest/rdf.yaml",
    "deepimagej/SkinLesionClassification/latest/rdf.yaml",
    "deepimagej/smlm-deepimagej/latest/rdf.yaml",
    "deepimagej/SMLMDensityMapEstimationDEFCoN/latest/rdf.yaml",
    "deepimagej/unet-pancreaticcellsegmentation/latest/rdf.yaml",
    "deepimagej/UNet2DGlioblastomaSegmentation/latest/rdf.yaml",
    "deepimagej/WidefieldDapiSuperResolution/latest/rdf.yaml",
    "deepimagej/WidefieldFitcSuperResolution/latest/rdf.yaml",
    "deepimagej/WidefieldTxredSuperResolution/latest/rdf.yaml",
    "dl4miceverywhere/DL4MicEverywhere/latest/rdf.yaml",
    "dl4miceverywhere/Notebook_bioimageio_pytorch/latest/rdf.yaml",
    "dl4miceverywhere/Notebook_bioimageio_tensorflow/latest/rdf.yaml",
    "fiji/Fiji/latest/rdf.yaml",
    "hpa/HPA-Classification/latest/rdf.yaml",
    "hpa/hpa-kaggle-2021-dataset/latest/rdf.yaml",
    "icy/icy/latest/rdf.yaml",
    "ilastik/arabidopsis_tissue_atlas/latest/rdf.yaml",
    "ilastik/cremi_training_data/latest/rdf.yaml",
    "ilastik/ilastik/latest/rdf.yaml",
    "ilastik/isbi2012_neuron_segmentation_challenge/latest/rdf.yaml",
    "ilastik/mitoem_segmentation_challenge/latest/rdf.yaml",
    "ilastik/mws-segmentation/latest/rdf.yaml",
    "imjoy/BioImageIO-Packager/latest/rdf.yaml",
    "imjoy/GenericBioEngineApp/latest/rdf.yaml",
    "imjoy/HPA-Single-Cell/latest/rdf.yaml",
    "imjoy/ImageJ.JS/latest/rdf.yaml",
    "imjoy/ImJoy/latest/rdf.yaml",
    "imjoy/LuCa-7color/latest/rdf.yaml",
    "imjoy/vizarr/latest/rdf.yaml",
    "qupath/QuPath/latest/rdf.yaml",
    "stardist/stardist/latest/rdf.yaml",
    "zero/Dataset_CARE_2D_coli_DeepBacs/latest/rdf.yaml",
    "zero/Dataset_CARE_2D_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Dataset_CARE_3D_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Dataset_CycleGAN_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Dataset_Deep-STORM_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Dataset_fnet_3D_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Dataset_fnet_DeepBacs/latest/rdf.yaml",
    "zero/Dataset_Noise2Void_2D_subtilis_DeepBacs/latest/rdf.yaml",
    "zero/Dataset_Noise2Void_2D_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Dataset_Noise2Void_3D_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Dataset_Noisy_Nuclei_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Dataset_pix2pix_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Dataset_SplineDist_2D_DeepBacs/latest/rdf.yaml",
    "zero/Dataset_StarDist_2D_DeepBacs/latest/rdf.yaml",
    "zero/Dataset_StarDist_2D_ZeroCostDL4Mic_2D/latest/rdf.yaml",
    "zero/Dataset_StarDist_brightfield_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Dataset_StarDist_brightfield2_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Dataset_StarDist_Fluo_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Dataset_StarDist_fluo2_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Dataset_U-Net_2D_DeepBacs/latest/rdf.yaml",
    "zero/Dataset_U-Net_2D_multilabel_DeepBacs/latest/rdf.yaml",
    "zero/Dataset_YOLOv2_antibiotic_DeepBacs/latest/rdf.yaml",
    "zero/Dataset_YOLOv2_coli_DeepBacs/latest/rdf.yaml",
    "zero/Dataset_YOLOv2_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Notebook Preview/latest/rdf.yaml",
    "zero/Notebook_Augmentor_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Notebook_CARE_2D_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Notebook_CARE_3D_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Notebook_Cellpose_2D_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Notebook_CycleGAN_2D_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Notebook_CycleGAN_2D_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Notebook_DecoNoising_2D_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Notebook_DecoNoising_2D_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Notebook_Deep-STORM_2D_ZeroCostDL4Mic_DeepImageJ/latest/rdf.yaml",
    "zero/Notebook_Deep-STORM_2D_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Notebook_DenoiSeg_2D_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Notebook_Detectron2_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Notebook_Detectron2_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Notebook_DFCAN_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Notebook_DRMIME_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Notebook_DRMIME_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Notebook_EmbedSeg_2D_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Notebook_EmbedSeg_2D_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Notebook_fnet_2D_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Notebook_fnet_3D_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Notebook_Interactive_Segmentation_Kaibu_2D_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Notebook_MaskRCNN_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Notebook_MaskRCNN_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Notebook_Noise2Void_2D_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Notebook_Noise2Void_3D_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Notebook_pix2pix_2D_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Notebook_pix2pix_2D_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/notebook_preview/latest/rdf.yaml-latest",
    "zero/notebook_preview/latest/rdf.yaml",
    "zero/Notebook_Quality_Control_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Notebook_RCAN_3D_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Notebook_RetinaNet_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Notebook_RetinaNet_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Notebook_SplineDist_2D_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Notebook_StarDist_2D_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Notebook_StarDist_3D_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Notebook_StarDist_3D_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Notebook_U-Net_2D_multilabel_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Notebook_U-Net_2D_multilabel_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Notebook_U-Net_2D_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Notebook_U-Net_2D_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Notebook_U-Net_3D_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Notebook_U-Net_3D_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/Notebook_YOLOv2_ZeroCostDL4Mic/latest/rdf.yaml",
    "zero/WGAN_ZeroCostDL4Mic.ipynb/latest/rdf.yaml",
}


def parse_args():
    p = ArgumentParser(
        description=(
            "report why some RDFs in the bioimage.io collection fail validation"
        )
    )
    _ = p.add_argument(
        "output", type=str, nargs="?", default="invalid_rdfs_{version}.md"
    )
    _ = p.add_argument("limit", type=int, nargs="?", default=200)
    _ = p.add_argument(
        "--version", default="discover", nargs="?", choices=["discover", "latest"]
    )
    args = p.parse_args()
    return args


def main(output: Path, limit: int, version: Literal["discover", "latest"]):
    if version == "discover":
        invalid = KNOWN_INVALID
    elif version == "latest":
        invalid = KNOWN_INVALID_AS_LATEST
    else:
        assert_never(version)

    invalid = [RDF_BASE_URL + k for k in sorted(invalid)[:limit]]

    summaries = [load_description_and_validate_format_only(rdf) for rdf in invalid]
    formatted = [s.format() for s in summaries]
    out = "\n\n".join(formatted)
    _ = output.write_text(out, encoding="utf-8")
    print(out)


if __name__ == "__main__":
    args = parse_args()
    main(Path(args.output.format(version=args.version)), args.limit, args.version)



import collections.abc
import string

from .._internal.io import BioimageioYamlContent
from ._v0_2_converter import convert_from_older_format as convert_from_older_format_v0_2


def convert_from_older_format(data: BioimageioYamlContent) -> None:
    """convert raw RDF data of an older format where possible"""
    # check if we have future format version
    fv = data.get("format_version", "0.2.0")
    if (
        not isinstance(fv, str)
        or fv.count(".") != 2
        or tuple(map(int, fv.split(".")[:2])) > (0, 3)
    ):
        return

    convert_from_older_format_v0_2(data)

    convert_attachments(data)

    _ = data.pop("download_url", None)
    _ = data.pop("rdf_source", None)

    if "name" in data and isinstance(data["name"], str):
        data["name"] = "".join(
            c if c in string.ascii_letters + string.digits + "_- ()" else " "
            for c in data["name"]
        )[:128]

    data["format_version"] = "0.3.0"


def convert_attachments(data: BioimageioYamlContent) -> None:
    a = data.get("attachments")
    if isinstance(a, collections.abc.Mapping):
        data["attachments"] = tuple({"source": file} for file in a.get("files", []))  # type: ignore



import io
import shutil
from pathlib import Path
from tempfile import TemporaryDirectory

from bioimageio.spec.model import v0_5


def test_save_bioimageio_package(unet2d_path: Path):
    from bioimageio.spec import save_bioimageio_package

    package_path = save_bioimageio_package(unet2d_path)
    assert package_path.exists()


def test_save_bioimageio_package_to_stream(unet2d_path: Path):
    from bioimageio.spec import save_bioimageio_package_to_stream

    output_stream = io.BytesIO()
    stream = save_bioimageio_package_to_stream(unet2d_path, output_stream=output_stream)
    assert isinstance(stream, io.BytesIO)
    assert stream == output_stream


def test_save_bioimageio_package_to_stream_default(unet2d_path: Path):
    from bioimageio.spec import save_bioimageio_package_to_stream

    stream = save_bioimageio_package_to_stream(unet2d_path)
    assert isinstance(stream, io.BytesIO)


def test_save_bioimageio_package_as_folder(unet2d_path: Path):
    from bioimageio.spec import load_description, save_bioimageio_package_as_folder

    with TemporaryDirectory() as tmp_dir:
        tmp_dir = Path(tmp_dir)

        package_folder = tmp_dir / "package"
        _ = save_bioimageio_package_as_folder(unet2d_path, output_path=package_folder)

        # load package
        model = load_description(package_folder)
        assert isinstance(model, v0_5.ModelDescr)

        # alter package
        doc = model.documentation
        assert isinstance(doc, v0_5.RelativeFilePath)
        new_doc = f"copy_{doc}"
        shutil.move(str(package_folder / str(doc)), str(package_folder / new_doc))
        model.documentation = package_folder / new_doc

        # export altered package
        altered_package = tmp_dir / "altered_package"
        altered_package = save_bioimageio_package_as_folder(
            model, output_path=altered_package, weights_priority_order=["onnx"]
        )
        assert (altered_package / new_doc).exists(), altered_package / new_doc

        # load altered package
        reloaded_model = load_description(altered_package)
        assert isinstance(reloaded_model, v0_5.ModelDescr)
        assert str(reloaded_model.documentation).startswith("copy_")



from typing import Literal, Optional

from .._internal.common_nodes import Node
from .._internal.io import FileDescr as FileDescr
from .._internal.io import Sha256 as Sha256
from .._internal.io_basics import AbsoluteFilePath as AbsoluteFilePath
from .._internal.url import HttpUrl as HttpUrl
from ..generic.v0_3 import VALID_COVER_IMAGE_EXTENSIONS as VALID_COVER_IMAGE_EXTENSIONS
from ..generic.v0_3 import Author as Author
from ..generic.v0_3 import BadgeDescr as BadgeDescr
from ..generic.v0_3 import CiteEntry as CiteEntry
from ..generic.v0_3 import Doi as Doi
from ..generic.v0_3 import GenericDescrBase
from ..generic.v0_3 import LinkedResource as LinkedResource
from ..generic.v0_3 import Maintainer as Maintainer
from ..generic.v0_3 import OrcidId as OrcidId
from ..generic.v0_3 import RelativeFilePath as RelativeFilePath
from ..generic.v0_3 import ResourceId as ResourceId
from ..generic.v0_3 import Uploader as Uploader
from ..generic.v0_3 import Version as Version
from .v0_2 import NotebookSource as NotebookSource


class NotebookId(ResourceId):
    pass


class NotebookDescr(GenericDescrBase, title="bioimage.io notebook specification"):
    """Bioimage.io description of a Jupyter notebook."""

    type: Literal["notebook"] = "notebook"

    id: Optional[NotebookId] = None
    """Model zoo (bioimage.io) wide, unique identifier (assigned by bioimage.io)"""

    parent: Optional[NotebookId] = None
    """The description from which this one is derived"""

    source: NotebookSource
    """The Jupyter notebook"""


class LinkedNotebook(Node):
    """Reference to a bioimage.io notebook."""

    id: NotebookId
    """A valid notebook `id` from the bioimage.io collection."""

    version_number: int
    """version number (n-th published version, not the semantic version) of linked notebook"""



from __future__ import annotations

from pathlib import Path, PurePath
from typing import Any, Dict, Union

import pytest
from pydantic import (
    FilePath,
    TypeAdapter,
    ValidationError,
)
from typing_extensions import Annotated, assert_never

from bioimageio.spec._internal.io import (
    RelativeFilePath,
    WithSuffix,
    validate_suffix,
)
from bioimageio.spec._internal.io_basics import AbsoluteFilePath
from bioimageio.spec._internal.url import HttpUrl
from bioimageio.spec._internal.validation_context import ValidationContext
from bioimageio.spec._internal.warning_levels import WARNING
from bioimageio.spec.generic.v0_3 import DocumentationSource, GenericDescr
from tests.conftest import UNET2D_ROOT
from tests.utils import check_node

EXAMPLE_COM = "https://example.com/"
EXAMPLE_COM_FILE = "https://example.com/file"


@pytest.mark.parametrize(
    "kwargs",
    [
        dict(
            authors=[{"name": "Me"}],
            cite=[dict(text="lala", url=EXAMPLE_COM)],
            description="the description",
            format_version=GenericDescr.implemented_format_version,
            license="BSD-2-Clause-FreeBSD",
            name="my name",
            type="my_type",
            unknown_extra_field="present",
            version="1",
        ),
        dict(
            attachments={"files": [EXAMPLE_COM_FILE], "something": 42},
            authors=[{"name": "Me"}],
            cite=[dict(text="lala", url=EXAMPLE_COM)],
            description="my description",
            format_version=GenericDescr.implemented_format_version,
            license="BSD-2-Clause-FreeBSD",
            name="your name",
            type="my_type",
            version="2",
        ),
    ],
)
def test_generic_valid(kwargs: Dict[str, Any]):
    check_node(GenericDescr, kwargs, context=ValidationContext(perform_io_checks=False))


@pytest.mark.parametrize(
    "kwargs,context",
    [
        pytest.param(
            dict(
                format_version=GenericDescr.implemented_format_version,
                name="my name",
                description="my description",
                authors=[{"name": "Me"}],
                type="my_type",
                version="0.1.0",
                license="BSD-2-Clause-FreeBSD",
                cite=[dict(text="lala", url=EXAMPLE_COM)],
            ),
            ValidationContext(warning_level=WARNING, perform_io_checks=False),
            id="deprecated license",
        ),
        (
            dict(
                format_version=GenericDescr.implemented_format_version,
                version="0.1.0",
                type="my_type",
                name="their name",
            ),
            ValidationContext(perform_io_checks=False),
        ),
        (
            dict(
                format_version=GenericDescr.implemented_format_version,
                version="0.1.0",
                type="my_type",
                name="its name",
                attachments={"files": [Path(__file__), "missing"], "something": 42},
            ),
            ValidationContext(perform_io_checks=False),
        ),
    ],
)
def test_generic_invalid(kwargs: Dict[str, Any], context: ValidationContext):
    check_node(GenericDescr, kwargs, context=context, is_invalid=True)


# @pytest.mark.parametrize("src", [UNET2D_ROOT / "README.md", text_md_url])
def test_documentation_source():
    from bioimageio.spec.generic.v0_3 import DocumentationSource

    doc_src = "https://example.com/away.md"
    adapter = TypeAdapter(DocumentationSource)
    with ValidationContext(perform_io_checks=False):
        valid = adapter.validate_python(doc_src)

    assert str(valid) == doc_src


def test_documentation_source_abs_path():
    from bioimageio.spec.generic.v0_3 import DocumentationSource

    doc_src = UNET2D_ROOT / "README.md"
    assert doc_src.exists(), doc_src
    adapter = TypeAdapter(DocumentationSource)

    valid = adapter.validate_python(doc_src)
    assert str(valid) == str(doc_src)

    data = adapter.dump_python(valid, mode="python")
    assert str(data) == str(doc_src)
    data = adapter.dump_python(valid, mode="json")
    assert str(data) == str(doc_src)

    doc_src = UNET2D_ROOT / "does_not_exist.md"
    assert not doc_src.exists(), doc_src
    with pytest.raises(ValidationError):
        _ = adapter.validate_python(doc_src)


with ValidationContext(perform_io_checks=False):
    text_md_url = HttpUrl("https://example.com/text.md")


def validate_md_suffix(value: Union[AbsoluteFilePath, RelativeFilePath, HttpUrl]):
    return validate_suffix(value, ".md", case_sensitive=True)


@pytest.mark.parametrize(
    "src,adapter",
    [
        (UNET2D_ROOT / "README.md", a)
        for a in [
            TypeAdapter(
                Annotated[FilePath, WithSuffix(".md", case_sensitive=True)]
            ),  # pyright: ignore[reportCallIssue]
            TypeAdapter(
                Annotated[Path, WithSuffix(".md", case_sensitive=True)]
            ),  # pyright: ignore[reportCallIssue]
            TypeAdapter(
                Annotated[PurePath, WithSuffix(".md", case_sensitive=True)]
            ),  # pyright: ignore[reportCallIssue]
            TypeAdapter(
                Annotated[
                    Union[PurePath, HttpUrl], WithSuffix(".md", case_sensitive=True)
                ]
            ),
            TypeAdapter(
                Annotated[
                    Union[AbsoluteFilePath, RelativeFilePath, HttpUrl],
                    WithSuffix(".md", case_sensitive=True),
                ]
            ),
            TypeAdapter(DocumentationSource),
            TypeAdapter(
                Annotated[DocumentationSource, WithSuffix(".md", case_sensitive=True)]
            ),
        ]
    ]
    + [
        (text_md_url, a)
        for a in [
            TypeAdapter(
                Annotated[HttpUrl, WithSuffix(".md", case_sensitive=True)]
            ),  # pyright: ignore[reportCallIssue]
            TypeAdapter(
                Annotated[
                    Union[PurePath, HttpUrl], WithSuffix(".md", case_sensitive=True)
                ]
            ),
            TypeAdapter(
                Annotated[
                    Union[AbsoluteFilePath, RelativeFilePath, HttpUrl],
                    WithSuffix(".md", case_sensitive=True),
                ]
            ),
            TypeAdapter(DocumentationSource),
            TypeAdapter(
                Annotated[DocumentationSource, WithSuffix(".md", case_sensitive=True)]
            ),
        ]
    ],
)
def test_with_suffix(src: Union[Path, HttpUrl], adapter: TypeAdapter[Any]):
    with ValidationContext(perform_io_checks=False):
        valid = adapter.validate_python(src)

    assert isinstance(valid, type(src))

    # dump_python(..., mode="python") returns RootModel's as is for some TypeAdapters based on Unions
    # see https://github.com/pydantic/pydantic/issues/8963
    # obj = adapter.dump_python(valid, mode="python")
    # if isinstance(src, Path):
    #     assert obj == src
    # elif isinstance(src, HttpUrl):
    #     assert obj == str(src)
    # else:
    #     assert_never(src)

    json_obj = adapter.dump_python(valid, mode="json")
    if isinstance(src, Path):
        assert json_obj == str(src)
    elif isinstance(src, HttpUrl):
        assert json_obj == str(src)
    else:
        assert_never(src)



import logging
import re

import pytest

from bioimageio.spec._internal.warning_levels import ALERT, INFO, WARNING


@pytest.mark.parametrize("unit", ["lx·s", "kg/m^2·s^-2"])
def test_si_unit(unit: str):
    from bioimageio.spec._internal.constants import SI_UNIT_REGEX

    assert re.fullmatch(SI_UNIT_REGEX, unit)


@pytest.mark.parametrize("unit", ["lxs", " kg", "kg/m^-2"])
def test_invalid_si_uni(unit: str):
    from bioimageio.spec._internal.constants import SI_UNIT_REGEX

    assert not re.fullmatch(SI_UNIT_REGEX, unit)


def test_warning_levels():
    """test that our warning levels nicely match python's logging module's levels"""
    assert ALERT < logging.ERROR
    assert ALERT > logging.WARNING
    assert WARNING == logging.WARNING
    assert INFO == logging.INFO


def test_known_gh_users_are_lowercase():
    from bioimageio.spec._internal.constants import (
        KNOWN_GH_USERS,
        KNOWN_INVALID_GH_USERS,
    )

    assert KNOWN_GH_USERS == {user.lower() for user in KNOWN_GH_USERS}
    assert KNOWN_INVALID_GH_USERS == {user.lower() for user in KNOWN_INVALID_GH_USERS}



from typing import Any, ClassVar, Type, Union

import pydantic
import requests
import requests.exceptions
from pydantic import AfterValidator, RootModel
from typing_extensions import Annotated

from .field_warning import issue_warning
from .root_url import RootHttpUrl
from .validation_context import validation_context_var


def _validate_url(url: Union[str, pydantic.HttpUrl]) -> pydantic.AnyUrl:
    url = str(url)
    if not validation_context_var.get().perform_io_checks:
        return pydantic.AnyUrl(url)

    if url.startswith("https://colab.research.google.com/github/"):
        # head request for colab returns "Value error, 405: Method Not Allowed"
        # therefore we check if the source notebook exists at github instead
        val_url = url.replace(
            "https://colab.research.google.com/github/", "https://github.com/"
        )
    else:
        val_url = url

    try:
        response = requests.head(val_url, timeout=(3, 3))
    except (
        requests.exceptions.ChunkedEncodingError,
        requests.exceptions.ContentDecodingError,
        requests.exceptions.InvalidHeader,
        requests.exceptions.InvalidJSONError,
        requests.exceptions.InvalidSchema,
        requests.exceptions.InvalidURL,
        requests.exceptions.MissingSchema,
        requests.exceptions.StreamConsumedError,
        requests.exceptions.TooManyRedirects,
        requests.exceptions.UnrewindableBodyError,
        requests.exceptions.URLRequired,
    ) as e:
        raise ValueError(
            f"Invalid URL '{url}': {e}\nrequest: {e.request}\nresponse: {e.response}"
        )
    except requests.RequestException as e:
        issue_warning(
            "Failed to validate URL '{value}': {error}\nrequest: {request}\nresponse: {response}",
            value=url,
            msg_context={"error": str(e), "response": e.response, "request": e.request},
        )
    except Exception as e:
        issue_warning(
            "Failed to validate URL '{value}': {error}",
            value=url,
            msg_context={"error": str(e)},
        )
    else:
        follow_up_with_get = False
        if response.status_code == 302:  # found
            pass
        elif response.status_code in (301, 303, 308):
            issue_warning(
                "URL redirected ({status_code}): consider updating {value} with new"
                + " location: {location}",
                value=url,
                msg_context={
                    "status_code": response.status_code,
                    "location": response.headers.get("location"),
                },
            )
        elif response.status_code == 403:  # forbidden
            follow_up_with_get = True
            issue_warning(
                "{status_code}: {reason} {value}",
                value=url,
                msg_context={
                    "status_code": response.status_code,
                    "reason": response.reason,
                },
            )
        elif response.status_code == 405:
            issue_warning(
                "{status_code}: {reason} {value}",
                value=url,
                msg_context={
                    "status_code": response.status_code,
                    "reason": response.reason,
                },
            )
        elif response.status_code != 200:
            raise ValueError(f"{response.status_code}: {response.reason} {url}")

        if follow_up_with_get:
            pass
            # TODO follow up forbidden head request with get
            # motivating example: 403: Forbidden https://elifesciences.org/articles/57613

    return pydantic.AnyUrl(url)


class HttpUrl(RootHttpUrl):
    root_model: ClassVar[Type[RootModel[Any]]] = RootModel[
        Annotated[pydantic.HttpUrl, AfterValidator(_validate_url)]
    ]



from __future__ import annotations

import shutil
from argparse import ArgumentParser
from collections import OrderedDict
from dataclasses import dataclass, field
from pathlib import Path
from pprint import pformat
from types import ModuleType
from typing import Any, Dict, Iterator, List, Optional, Tuple, Type, get_args

from pydantic.alias_generators import to_pascal, to_snake
from pydantic.fields import FieldInfo
from pydantic_core import PydanticUndefined

from bioimageio.spec import (
    ResourceDescr,
    application,
    collection,
    dataset,
    generic,
    model,
    notebook,
)
from bioimageio.spec._internal.common_nodes import Node
from bioimageio.spec._internal.constants import IN_PACKAGE_MESSAGE
from bioimageio.spec._internal.utils import unindent

Loc = Tuple[str, ...]

ANNOTATION_MAP = {
    "pydantic_core._pydantic_core.Url": "Url",
    "typing.": "",
    "pathlib.": "",
    "bioimageio.spec._internal.common_nodes.FrozenDictNode": "Dict",
    "bioimageio.spec._internal.common_nodes.Kwargs": "Dict",
    "bioimageio.spec.types.": "",
    "pydantic.networks.EmailStr": "Email",
    "bioimageio.spec.": "",
    "NoneType": "None",
    "Ellipsis": "...",
    "PathType(path_type='dir')": "Directory",
}
MAX_LINE_WIDTH = 120

ADDITIONAL_DESCRIPTION_ANY_RESOURCE = (
    "\n**General notes on this documentation:**\n"
    "| symbol | explanation |\n"
    "| --- | --- |\n"
    "| `field`<sub>type hint</sub> | A fields's <sub>expected type</sub> may be"
    " shortened. "
    "If so, the abbreviated or full type is displayed below the field's description and"
    " can expanded to view "
    "further (nested) details if available. |\n"
    "| Union[A, B, ...] | indicates that a field value may be of type A or B, etc.|\n"
    "| Literal[a, b, ...] | indicates that a field value must be the specific value a"
    " or b, etc.|\n"
    "| Type* := Type (restrictions) | A field Type* followed by an asterisk indicates"
    " that annotations, e.g. "
    "value restriction apply. These are listed in parentheses in the expanded type"
    " description. "
    "They are not always intuitively understandable and merely a hint at more complex"
    " validation.|\n"
    r"| \<type\>.v\<major\>_\<minor\>.\<sub spec\> | "
    "Subparts of a spec might be taken from another spec type or format version. |\n"
    "| `field` ≝ `default` | Default field values are indicated after '≝' and make a"
    " field optional. "
    "However, `type` and `format_version` alwyas need to be set for resource"
    " descriptions written as YAML files "
    "and determine which bioimage.io specification applies. They are optional only when"
    " creating a resource "
    "description in Python code using the appropriate, `type` and `format_version`"
    " specific class.|\n"
    "| `field` ≝ 🡇 | Default field value is not displayed in-line, but in the code"
    " block below. |\n"
    f"| {IN_PACKAGE_MESSAGE} | Files referenced in fields which are marked with"
    f" '{IN_PACKAGE_MESSAGE}' "
    "are included when packaging the resource to a .zip archive. "
    "The resource description YAML file (RDF) is always included well as"
    " 'rdf.yaml'. |\n"
)


def anchor_tag(heading: str):
    a = heading.strip().strip("#")
    for rm in ",;!?./<>=`'\"":
        a = a.replace(rm, "")

    return "#" + a.replace(" ", "-")


def get_subnodes(loc: Loc, annotation: Any) -> Iterator[Tuple[Loc, Type[Node]]]:
    try:
        is_node = issubclass(annotation, Node)
    except TypeError:
        is_node = False

    if is_node:
        yield loc, annotation
    else:
        for like_list in ["List", "Tuple", "Set"]:
            if str(annotation).startswith(f"typing.{like_list}["):
                loc = loc[:-1] + (loc[-1] + ".i",)
                break

        for sa in get_args(annotation):
            yield from get_subnodes(loc, sa)


@dataclass
class AnnotationName:
    annotation: Any
    indent_level: int
    footnotes: OrderedDict[str, str]
    full_maybe_multiline: str = field(init=False)
    full_inline: str = field(init=False)
    abbreviated: Optional[str] = field(init=False)
    kind: str = field(init=False)

    annotation_map: Dict[str, str]

    def __post_init__(self):
        self.full_maybe_multiline = self.get_name(
            self.annotation, abbreviate=False, inline=False
        )
        self.full_inline = self.get_name(self.annotation, abbreviate=False)
        self.kind = self._get_kind()
        if self.indent_level + len(self.full_inline) > MAX_LINE_WIDTH:
            self.abbreviated = self.get_name(self.annotation, abbreviate=True)
        else:
            self.abbreviated = None

    def _get_kind(self):
        s = self.full_inline
        brackets = 0
        max_balance = -1
        for i in range(min(len(s), 32)):
            if s[i] == "[":
                brackets += 1

            if s[i] == "]":
                brackets -= 1

            if brackets == 0:
                max_balance = i

        return s[: max_balance + 1]

    def slim(self, s: str):
        """shortening that's always OK"""
        s = s.strip("'\"")
        if s.startswith("<class ") and s.endswith(">"):
            s = s[len("<class ") : -1]

        s = s.strip("'\"")
        for k, v in self.annotation_map.items():
            if s.startswith(k):
                s = v + s[len(k) :]
                break

        return s.strip("'\"")

    def more_common_sequence_name(self, type_name: str):
        bracket = type_name.find("[")
        if bracket == -1:
            first_part = type_name
        else:
            first_part = type_name[:bracket]

        common_name = {"List": "Sequence", "Tuple": "Sequence"}.get(
            first_part, first_part
        )
        if bracket == -1:
            return common_name
        else:
            return common_name + type_name[bracket:]

    def get_name(
        self, t: Any, abbreviate: bool, inline: bool = True, multiline_level: int = 0
    ) -> str:
        if isinstance(t, FieldInfo):
            parts = list(t.metadata)
            if t.discriminator:
                parts.append(f"discriminator={t.discriminator}")

            return "; ".join(parts)

        s = self.slim(str(t))
        if s.startswith("Annotated["):
            args = get_args(t)
            if abbreviate:
                return f"{self.get_name(args[0], abbreviate, inline, multiline_level)}*"

            annotated_type = self.get_name(args[0], abbreviate, inline, multiline_level)
            annos = f"({'; '.join([self.get_name(tt, abbreviate, inline, multiline_level) for tt in args[1:]])})"
            if (
                inline
                or abbreviate
                or (
                    multiline_level + len(annotated_type) + 1 + len(annos)
                    < MAX_LINE_WIDTH
                )
            ):
                anno_sep = " "
            else:
                anno_sep = "\n" + " " * multiline_level * 2

            return f"{annotated_type}{anno_sep}{annos}"

        if s.startswith("Optional["):
            return f"Optional[{self.get_name(get_args(t)[0], abbreviate, inline, multiline_level)}]"

        for format_like_seq in ["Union", "Tuple", "Literal", "Dict", "List", "Set"]:
            if not s.startswith(format_like_seq):
                continue

            args = get_args(t)
            if format_like_seq == "Tuple" and len(args) == 2 and args[1] == ...:
                args = args[:1]

            format_like_seq_name = self.more_common_sequence_name(format_like_seq)

            if len(args) > 4 and abbreviate:
                args = [args[0], "...", args[-1]]

            parts = [
                self.get_name(tt, abbreviate, inline, multiline_level) for tt in args
            ]
            one_line = f"{format_like_seq_name}[{', '.join(parts)}]"
            if (
                abbreviate
                or inline
                or (self.indent_level + len(one_line) < MAX_LINE_WIDTH)
            ):
                return one_line

            first_line_descr = f"{format_like_seq_name} of"
            if len(args) == 1:
                more_maybe_multiline = self.get_name(
                    args[0],
                    abbreviate=abbreviate,
                    inline=inline,
                    multiline_level=multiline_level,
                )
                return first_line_descr + " " + more_maybe_multiline

            parts = [
                self.get_name(
                    tt, abbreviate, inline=inline, multiline_level=multiline_level + 1
                )
                for tt in args
            ]
            multiline_parts = f"\n{' '* multiline_level * 2}- ".join(parts)
            return (
                f"{first_line_descr}\n{' '* multiline_level * 2}- {multiline_parts}\n"
            )

        return s


class Field:
    STYLE_SWITCH_DEPTH = 4

    def __init__(
        self,
        loc: Loc,
        info: FieldInfo,
        *,
        footnotes: OrderedDict[str, str],
        rd_class: type[ResourceDescr],
        all_examples: List[Tuple[str, List[Any]]],
    ) -> None:
        super().__init__()
        assert loc
        self.loc = loc
        self.info = info
        self.footnotes = footnotes
        self.annotation_map = {f"{rd_class.__module__}.": "", **ANNOTATION_MAP}
        self.rd_class = rd_class
        self.all_examples = all_examples

    @property
    def indent_with_symbol(self):
        spaces = " " * max(0, self.indent_level - 2)
        if len(self.loc) <= self.STYLE_SWITCH_DEPTH:
            symbol = f"#{'#'* len(self.loc)} "
        else:
            symbol = "* "

        return f"{spaces}{symbol}"

    @property
    def indent_level(self):
        return max(0, len(self.loc) - self.STYLE_SWITCH_DEPTH) * 2

    @property
    def indent_spaces(self):
        return " " * self.indent_level

    @property
    def name(self):
        n = ".".join(self.loc)
        if len(self.loc) <= self.STYLE_SWITCH_DEPTH:
            return f"`{n}`"
        else:
            return f'<a id="{n}"></a>`{n}`'

    def get_explanation(self):
        title = self.info.title or ""
        description = unindent(self.info.description or "", ignore_first_line=True)
        ret = self.indent_spaces
        if title:
            ret += f"{title}: "
            if "\n" in description or len(ret) + len(description) > MAX_LINE_WIDTH:
                ret += "\n"

        ret += description.strip() + "\n"

        if self.info.examples:
            ex = "Example" if len(self.info.examples) == 1 else "Examples"
            ex = f"*{ex}:*"
            if len(self.info.examples) == 1:
                e = self.info.examples[0]
                example_inline = f"'{e}'" if isinstance(e, str) else str(e)
            else:
                example_inline = str(self.info.examples)
                if self.indent_level + len(example_inline) > MAX_LINE_WIDTH:
                    for i in range(len(self.info.examples) - 1, 0, -1):
                        example_inline = str(self.info.examples[:i] + ["…"])
                        if self.indent_level + len(example_inline) <= MAX_LINE_WIDTH:
                            break

            ret += f"[{ex}]({anchor_tag(self.name)}) {example_inline}\n"
            self.all_examples.append((self.name, self.info.examples))

        return ret.replace("\n", self.indent_spaces + "\n")

    def get_default_value(self):
        d = self.info.get_default(call_default_factory=True)
        if d is PydanticUndefined:
            return ""
        # elif d == "":
        #     d = "<empty string>"
        d_inline = f"`{d}`"
        if self.indent_level + 30 + len(d_inline) > MAX_LINE_WIDTH:
            return f" ≝ 🡇\n```python\n{pformat(d, indent=self.indent_level, width=MAX_LINE_WIDTH)}\n```\n"
        else:
            return f" ≝ {d_inline}"

    def get_md(self) -> str:
        nested = ""
        for subloc, subnode in get_subnodes(self.loc, self.info.annotation):
            sub_anno = AnnotationName(
                annotation=subnode,
                footnotes=self.footnotes,
                indent_level=self.indent_level + 2,
                annotation_map=self.annotation_map,
            ).full_inline
            subfields = ""
            for sfn, sinfo in subnode.model_fields.items():
                subfields += (
                    "\n"
                    + Field(
                        subloc + (sfn,),
                        sinfo,
                        footnotes=self.footnotes,
                        rd_class=self.rd_class,
                        all_examples=self.all_examples,
                    ).get_md()
                )
            if subfields:
                nested += f"\n{self.indent_spaces}**{sub_anno}:**{subfields}"

        an = AnnotationName(
            annotation=self.info.annotation,
            footnotes=self.footnotes,
            indent_level=self.indent_level,
            annotation_map=self.annotation_map,
        )
        first_line = (
            f"{self.indent_with_symbol}{self.name}<sub>"
            f" {an.kind}</sub>{self.get_default_value()}\n"
        )
        if (nested or an.abbreviated) and len(self.loc) <= self.STYLE_SWITCH_DEPTH:
            if an.abbreviated is None:
                expaned_type_anno = ""
            else:
                expaned_type_anno = an.full_maybe_multiline + "\n"

            ret = (
                f"{first_line}{self.get_explanation()}\n"
                f"<details><summary>{an.abbreviated or an.full_inline}\n\n</summary>\n\n"
                f"{expaned_type_anno}{nested}\n</details>\n"
            )
        else:
            if an.kind == an.full_inline:
                expaned_type_anno = ""
            else:
                expaned_type_anno = "\n" + an.full_inline

            ret = f"{first_line}{self.get_explanation()}\n{expaned_type_anno}{nested}\n"

        return ret


def get_documentation_file_name(
    rd_class: Type[ResourceDescr], *, latest: bool = False, minor: bool = False
):
    assert not (latest and minor)
    typ = to_snake(rd_class.__name__)
    if latest:
        v = "latest"
    elif minor:
        v = "v" + "-".join(rd_class.implemented_format_version.split(".")[:2])
    else:
        v = f"v{rd_class.implemented_format_version.replace('.', '-')}"

    return f"{typ}_{v}.md"


def export_documentation(folder: Path, rd_class: Type[ResourceDescr]) -> Path:
    footnotes: OrderedDict[str, str] = OrderedDict()
    all_examples: List[Tuple[str, List[Any]]] = []
    md = (
        "# "
        + (rd_class.model_config.get("title") or "")
        + "\n"
        + (
            unindent(rd_class.__doc__ or "", ignore_first_line=True)
            + ADDITIONAL_DESCRIPTION_ANY_RESOURCE
        )
    )
    all_fields = [
        (fn, info)
        for fn, info in rd_class.model_fields.items()
        if fn not in ("type", "format_version")
    ]

    def field_sort_key(fn_info: Tuple[str, FieldInfo]) -> Tuple[bool, str]:
        fn, info = fn_info
        return (
            info.get_default(call_default_factory=True) is not PydanticUndefined,
            fn,
        )

    all_fields = sorted(all_fields, key=field_sort_key)

    field_names = ["type", "format_version"] + [fn for (fn, _) in all_fields]
    for field_name in field_names:
        info = rd_class.model_fields[field_name]
        md += (
            "\n"
            + Field(
                (field_name,),
                info,
                footnotes=footnotes,
                rd_class=rd_class,
                all_examples=all_examples,
            ).get_md()
        )

    md += "\n"
    for i, full in enumerate(footnotes, start=1):
        md += f"\n[^{i}]: {full}"

    if all_examples:
        md += "# Example values\n"
        for name, examples in all_examples:
            if len(examples) == 1:
                formatted_examples = str(examples[0])
            else:
                formatted_examples = "".join(f"- {ex}\n" for ex in examples)

            md += f"### {name}\n{formatted_examples}\n"

    if footnotes:
        md += "\n"

    for file_path in [
        folder / get_documentation_file_name(rd_class, minor=True),
        folder / get_documentation_file_name(rd_class),
    ]:
        _ = file_path.write_text(md, encoding="utf-8")
        print(f"written {file_path}")

    return file_path  # type: ignore


def export_module_documentations(folder: Path, module: ModuleType):
    rd_name = to_pascal(module.__name__.split(".")[-1]) + "Descr"

    rd_class = None
    latest = None
    v = None
    for v in sorted(dir(module)):
        v_module = getattr(module, v)
        if not hasattr(v_module, rd_name):
            continue

        rd_class = getattr(v_module, rd_name)
        latest = export_documentation(folder, rd_class)

    assert latest is not None
    assert rd_class is not None
    shutil.copy(
        str(latest), folder / get_documentation_file_name(rd_class, latest=True)
    )
    print(f" copied {latest} as latest")


def main(dist: Path):
    dist.mkdir(exist_ok=True, parents=True)

    export_module_documentations(dist, application)
    export_module_documentations(dist, collection)
    export_module_documentations(dist, dataset)
    export_module_documentations(dist, generic)
    export_module_documentations(dist, model)
    export_module_documentations(dist, notebook)


def parse_args():
    p = ArgumentParser(
        description="script that generates bioimageio user documentation"
    )
    _ = p.add_argument(
        "--dist",
        nargs="?",
        default=str((Path(__file__).parent / "../user_docs").resolve()),
    )
    args = p.parse_args()
    return args


if __name__ == "__main__":
    args = parse_args()
    main(dist=Path(args.dist))



from __future__ import annotations

from datetime import datetime
from keyword import iskeyword
from typing import Any, ClassVar, Sequence, Type, TypeVar, Union

import annotated_types
from dateutil.parser import isoparse
from pydantic import PlainSerializer, RootModel, StringConstraints
from typing_extensions import Annotated, Literal

from .constants import DOI_REGEX, SI_UNIT_REGEX
from .io import FileSource as FileSource
from .io import ImportantFileSource as ImportantFileSource
from .io import PermissiveFileSource as PermissiveFileSource
from .io import RelativeFilePath as RelativeFilePath
from .io import Sha256 as Sha256
from .io_basics import AbsoluteDirectory as AbsoluteDirectory
from .io_basics import AbsoluteFilePath as AbsoluteFilePath
from .io_basics import FileName as FileName
from .license_id import DeprecatedLicenseId as DeprecatedLicenseId
from .license_id import LicenseId as LicenseId
from .url import HttpUrl as HttpUrl
from .validated_string import ValidatedString
from .validator_annotations import AfterValidator, BeforeValidator
from .version_type import Version as Version

S = TypeVar("S", bound=Sequence[Any])
NotEmpty = Annotated[S, annotated_types.MinLen(1)]


def _validate_identifier(s: str) -> str:
    if not s.isidentifier():
        raise ValueError(
            f"'{s}' is not a valid (Python) identifier, see"
            + " https://docs.python.org/3/reference/lexical_analysis.html#identifiers"
            + " for details."
        )

    return s


def _validate_is_not_keyword(s: str) -> str:
    if iskeyword(s):
        raise ValueError(f"'{s}' is a Python keyword and not allowed here.")

    return s


def _validate_datetime(dt: Union[datetime, str, Any]) -> datetime:
    if isinstance(dt, datetime):
        return dt
    elif isinstance(dt, str):
        return isoparse(dt)

    raise ValueError(f"'{dt}' not a string or datetime.")


def _validate_orcid_id(orcid_id: str):
    if len(orcid_id) == 19 and all(orcid_id[idx] == "-" for idx in [4, 9, 14]):
        check = 0
        for n in orcid_id[:4] + orcid_id[5:9] + orcid_id[10:14] + orcid_id[15:]:
            # adapted from stdnum.iso7064.mod_11_2.checksum()
            check = (2 * check + int(10 if n == "X" else n)) % 11
        if check == 1:
            return orcid_id  # valid

    raise ValueError(
        f"'{orcid_id} is not a valid ORCID iD in hyphenated groups of 4 digits."
    )


# TODO follow up on https://github.com/pydantic/pydantic/issues/8964
# to remove _serialize_datetime
def _serialize_datetime_json(dt: datetime) -> str:
    return dt.isoformat()


class Datetime(
    RootModel[
        Annotated[
            datetime,
            BeforeValidator(_validate_datetime),
            PlainSerializer(_serialize_datetime_json, when_used="json-unless-none"),
        ]
    ]
):
    """Timestamp in [ISO 8601](#https://en.wikipedia.org/wiki/ISO_8601) format
    with a few restrictions listed [here](https://docs.python.org/3/library/datetime.html#datetime.datetime.fromisoformat).
    """


class Doi(ValidatedString):
    root_model: ClassVar[Type[RootModel[Any]]] = RootModel[
        Annotated[str, StringConstraints(pattern=DOI_REGEX)]
    ]


FormatVersionPlaceholder = Literal["latest", "discover"]
IdentifierAnno = Annotated[
    NotEmpty[str],
    AfterValidator(_validate_identifier),
    AfterValidator(_validate_is_not_keyword),
]


class Identifier(ValidatedString):
    root_model: ClassVar[Type[RootModel[Any]]] = RootModel[IdentifierAnno]


LowerCaseIdentifierAnno = Annotated[IdentifierAnno, annotated_types.LowerCase]


class LowerCaseIdentifier(ValidatedString):
    root_model: ClassVar[Type[RootModel[Any]]] = RootModel[LowerCaseIdentifierAnno]


class OrcidId(ValidatedString):
    """an ORCID identifier"""

    root_model: ClassVar[Type[RootModel[Any]]] = RootModel[
        Annotated[str, AfterValidator(_validate_orcid_id)]
    ]


class SiUnit(ValidatedString):
    """Si unit"""

    root_model: ClassVar[Type[RootModel[Any]]] = RootModel[
        Annotated[
            str,
            StringConstraints(min_length=1, pattern=SI_UNIT_REGEX),
            BeforeValidator(
                lambda s: (
                    s.replace("×", "·").replace("*", "·").replace(" ", "·")
                    if isinstance(s, str)
                    else s
                )
            ),
        ]
    ]



# autogen: start
"""
implementaions of all released minor versions are available in submodules:
- collection v0_2: `bioimageio.spec.collection.v0_2.CollectionDescr` [user documentation](../../../user_docs/collection_descr_v0-2.md)
- collection v0_3: `bioimageio.spec.collection.v0_3.CollectionDescr` [user documentation](../../../user_docs/collection_descr_v0-3.md)
"""
from typing import Union

from pydantic import Discriminator
from typing_extensions import Annotated

from .v0_2 import CollectionDescr as CollectionDescr_v0_2
from .v0_3 import CollectionDescr as CollectionDescr
from .v0_3 import CollectionDescr as CollectionDescr_v0_3

AnyCollectionDescr = Annotated[
    Union[CollectionDescr_v0_2, CollectionDescr_v0_3], Discriminator("format_version")
]
"""Union of any released collection desription"""
# autogen: stop



import pytest

from bioimageio.spec._internal.io_basics import FileName


@pytest.mark.parametrize(
    "name",
    [
        "bioimageio.yaml",
        "rdf.yaml",
        "model.yaml",
        "smth.bioimageio.yaml",
        "smth.rdf.yaml",
        "smth.model.yaml",
    ],
)
def test_is_valid_rdf_name(name: FileName):
    from bioimageio.spec._internal.io import is_valid_bioimageio_yaml_name

    assert is_valid_bioimageio_yaml_name(name), name


@pytest.mark.parametrize("name", ["bioimageio.yml", "RDF.yaml", "smth.yaml"])
def test_is_invalid_rdf_name(name: FileName):
    from bioimageio.spec._internal.io import is_valid_bioimageio_yaml_name

    assert not is_valid_bioimageio_yaml_name(name), name



"""
.. include:: ../../README.md
"""

from . import application as application
from . import collection as collection
from . import dataset as dataset
from . import generic as generic
from . import model as model
from ._description import LatestResourceDescr as LatestResourceDescr
from ._description import ResourceDescr as ResourceDescr
from ._description import SpecificResourceDescr as SpecificResourceDescr
from ._description import build_description as build_description
from ._description import dump_description as dump_description
from ._description import validate_format as validate_format
from ._internal import settings as settings
from ._internal.common_nodes import InvalidDescr as InvalidDescr
from ._internal.constants import VERSION
from ._internal.validation_context import ValidationContext as ValidationContext
from ._io import load_description as load_description
from ._io import (
    load_description_and_validate_format_only as load_description_and_validate_format_only,
)
from ._io import save_bioimageio_yaml_only as save_bioimageio_yaml_only
from ._package import save_bioimageio_package as save_bioimageio_package
from ._package import (
    save_bioimageio_package_as_folder as save_bioimageio_package_as_folder,
)
from ._package import (
    save_bioimageio_package_to_stream as save_bioimageio_package_to_stream,
)
from .application import AnyApplicationDescr as AnyApplicationDescr
from .application import ApplicationDescr as ApplicationDescr
from .collection import AnyCollectionDescr as AnyCollectionDescr
from .collection import CollectionDescr as CollectionDescr
from .dataset import AnyDatasetDescr as AnyDatasetDescr
from .dataset import DatasetDescr as DatasetDescr
from .generic import AnyGenericDescr as AnyGenericDescr
from .generic import GenericDescr as GenericDescr
from .model import AnyModelDescr as AnyModelDescr
from .model import ModelDescr as ModelDescr
from .notebook import AnyNotebookDescr as AnyNotebookDescr
from .notebook import NotebookDescr as NotebookDescr
from .pretty_validation_errors import (
    enable_pretty_validation_errors_in_ipynb as enable_pretty_validation_errors_in_ipynb,
)
from .summary import ValidationSummary as ValidationSummary

__version__ = VERSION



from pathlib import Path

import pytest


def test_load_non_existing_rdf():
    from bioimageio.spec import load_description

    spec_path = Path("some/none/existing/path/to/spec.model.yaml")

    with pytest.raises(FileNotFoundError):
        _ = load_description(spec_path)



from __future__ import annotations

import string
from functools import partial
from typing import (
    Any,
    ClassVar,
    Dict,
    List,
    Literal,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
)

import annotated_types
from annotated_types import Len, LowerCase, MaxLen, MinLen
from pydantic import Field, RootModel, ValidationInfo, field_validator, model_validator
from typing_extensions import Annotated

from .._internal.common_nodes import (
    Converter,
    Node,
    ResourceDescrBase,
)
from .._internal.constants import (
    TAG_CATEGORIES,
)
from .._internal.field_validation import validate_gh_user
from .._internal.field_warning import as_warning, warn
from .._internal.io import (
    BioimageioYamlContent,
    V_suffix,
    YamlValue,
    include_in_package_serializer,
    validate_suffix,
)
from .._internal.io import FileDescr as FileDescr
from .._internal.io import Sha256 as Sha256
from .._internal.io_basics import AbsoluteFilePath
from .._internal.license_id import LicenseId
from .._internal.types import (
    DeprecatedLicenseId,
    ImportantFileSource,
    NotEmpty,
)
from .._internal.types import RelativeFilePath as RelativeFilePath
from .._internal.url import HttpUrl as HttpUrl
from .._internal.validated_string import ValidatedString
from .._internal.validator_annotations import (
    AfterValidator,
    Predicate,
    RestrictCharacters,
)
from .._internal.version_type import Version as Version
from .._internal.warning_levels import ALERT, INFO
from ._v0_3_converter import convert_from_older_format
from .v0_2 import VALID_COVER_IMAGE_EXTENSIONS, CoverImageSource
from .v0_2 import Author as _Author_v0_2
from .v0_2 import BadgeDescr as BadgeDescr
from .v0_2 import Doi as Doi
from .v0_2 import Maintainer as _Maintainer_v0_2
from .v0_2 import OrcidId as OrcidId
from .v0_2 import Uploader as Uploader

KNOWN_SPECIFIC_RESOURCE_TYPES = (
    "application",
    "collection",
    "dataset",
    "model",
    "notebook",
)


class ResourceId(ValidatedString):
    root_model: ClassVar[Type[RootModel[Any]]] = RootModel[
        Annotated[
            NotEmpty[str],
            RestrictCharacters(string.ascii_lowercase + string.digits + "_-/."),
            annotated_types.Predicate(
                lambda s: not (s.startswith("/") or s.endswith("/"))
            ),
        ]
    ]


def _validate_md_suffix(value: V_suffix) -> V_suffix:
    return validate_suffix(value, suffix=".md", case_sensitive=True)


DocumentationSource = Annotated[
    Union[AbsoluteFilePath, RelativeFilePath, HttpUrl],
    AfterValidator(_validate_md_suffix),
    include_in_package_serializer,
]


def _has_no_slash(s: str) -> bool:
    return "/" not in s and "\\" not in s


class Author(_Author_v0_2):
    name: Annotated[str, Predicate(_has_no_slash)]
    github_user: Optional[str] = None

    @field_validator("github_user", mode="after")
    def _validate_gh_user(cls, value: Optional[str]):
        if value is None:
            return None
        else:
            return validate_gh_user(value)


class _AuthorConv(Converter[_Author_v0_2, Author]):
    def _convert(
        self, src: _Author_v0_2, tgt: "type[Author] | type[dict[str, Any]]"
    ) -> "Author | dict[str, Any]":
        return tgt(
            name=src.name,
            github_user=src.github_user,
            affiliation=src.affiliation,
            email=src.email,
            orcid=src.orcid,
        )


_author_conv = _AuthorConv(_Author_v0_2, Author)


class Maintainer(_Maintainer_v0_2):
    name: Optional[Annotated[str, Predicate(_has_no_slash)]] = None
    github_user: str

    @field_validator("github_user", mode="after")
    def validate_gh_user(cls, value: str):
        return validate_gh_user(value)


class _MaintainerConv(Converter[_Maintainer_v0_2, Maintainer]):
    def _convert(
        self, src: _Maintainer_v0_2, tgt: "type[Maintainer | dict[str, Any]]"
    ) -> "Maintainer | dict[str, Any]":
        return tgt(
            name=src.name,
            github_user=src.github_user,
            affiliation=src.affiliation,
            email=src.email,
            orcid=src.orcid,
        )


_maintainer_conv = _MaintainerConv(_Maintainer_v0_2, Maintainer)


class CiteEntry(Node):
    text: str
    """free text description"""

    doi: Optional[Doi] = None
    """A digital object identifier (DOI) is the prefered citation reference.
    See https://www.doi.org/ for details. (alternatively specify `url`)"""

    url: Optional[HttpUrl] = None
    """URL to cite (preferably specify a `doi` instead)"""

    @model_validator(mode="after")
    def _check_doi_or_url(self):
        if not self.doi and not self.url:
            raise ValueError("Either 'doi' or 'url' is required")

        return self


class LinkedResource(Node):
    """Reference to a bioimage.io resource"""

    id: ResourceId
    """A valid resource `id` from the official bioimage.io collection."""

    version_number: int
    """version number (n-th published version, not the semantic version) of linked resource"""


class GenericModelDescrBase(ResourceDescrBase):
    """Base for all resource descriptions including of model descriptions"""

    name: Annotated[
        Annotated[
            str, RestrictCharacters(string.ascii_letters + string.digits + "_- ()")
        ],
        MinLen(5),
        MaxLen(128),
        warn(MaxLen(64), "Name longer than 64 characters.", INFO),
    ]
    name: Annotated[NotEmpty[str], MaxLen(128)]
    """A human-friendly name of the resource description.
    May only contains letters, digits, underscore, minus, parentheses and spaces."""

    description: Annotated[
        str, MaxLen(1024), warn(MaxLen(512), "Description longer than 512 characters.")
    ]
    """A string containing a brief description."""

    covers: Annotated[
        List[CoverImageSource],
        Field(
            examples=[],
            description=(
                "Cover images. Please use an image smaller than 500KB and an aspect"
                " ratio width to height of 2:1 or 1:1.\nThe supported image formats"
                f" are: {VALID_COVER_IMAGE_EXTENSIONS}"
            ),
        ),
    ] = Field(default_factory=list)
    """∈📦 Cover images."""

    id_emoji: Optional[Annotated[str, Len(min_length=1, max_length=1)]] = None
    """UTF-8 emoji for display alongside the `id`."""

    authors: NotEmpty[List[Author]]
    """The authors are the creators of this resource description and the primary points of contact."""

    attachments: List[FileDescr] = Field(default_factory=list)
    """file attachments"""

    cite: NotEmpty[List[CiteEntry]]
    """citations"""

    license: Annotated[
        Annotated[
            Union[LicenseId, DeprecatedLicenseId], Field(union_mode="left_to_right")
        ],
        warn(
            LicenseId,
            "{value} is deprecated, see https://spdx.org/licenses/{value}.html",
        ),
        Field(examples=["CC0-1.0", "MIT", "BSD-2-Clause"]),
    ]
    """A [SPDX license identifier](https://spdx.org/licenses/).
    We do not support custom license beyond the SPDX license list, if you need that please
    [open a GitHub issue](https://github.com/bioimage-io/spec-bioimage-io/issues/new/choose)
    to discuss your intentions with the community."""

    config: Annotated[
        Dict[str, YamlValue],
        Field(
            examples=[
                dict(
                    bioimageio={
                        "my_custom_key": 3837283,
                        "another_key": {"nested": "value"},
                    },
                    imagej={"macro_dir": "path/to/macro/file"},
                )
            ],
        ),
    ] = Field(default_factory=dict)
    """A field for custom configuration that can contain any keys not present in the RDF spec.
    This means you should not store, for example, a GitHub repo URL in `config` since there is a `git_repo` field.
    Keys in `config` may be very specific to a tool or consumer software. To avoid conflicting definitions,
    it is recommended to wrap added configuration into a sub-field named with the specific domain or tool name,
    for example:
    ```yaml
    config:
        bioimageio:  # here is the domain name
            my_custom_key: 3837283
            another_key:
                nested: value
        imagej:       # config specific to ImageJ
            macro_dir: path/to/macro/file
    ```
    If possible, please use [`snake_case`](https://en.wikipedia.org/wiki/Snake_case) for keys in `config`.
    You may want to list linked files additionally under `attachments` to include them when packaging a resource.
    (Packaging a resource means downloading/copying important linked files and creating a ZIP archive that contains
    an altered rdf.yaml file with local references to the downloaded files.)"""

    git_repo: Annotated[
        Optional[HttpUrl],
        Field(
            examples=[
                "https://github.com/bioimage-io/spec-bioimage-io/tree/main/example_descriptions/models/unet2d_nuclei_broad"
            ],
        ),
    ] = None
    """A URL to the Git repository where the resource is being developed."""

    icon: Union[
        Annotated[str, Len(min_length=1, max_length=2)], ImportantFileSource, None
    ] = None
    """An icon for illustration, e.g. on bioimage.io"""

    links: Annotated[
        List[str],
        Field(
            examples=[
                (
                    "ilastik/ilastik",
                    "deepimagej/deepimagej",
                    "zero/notebook_u-net_3d_zerocostdl4mic",
                )
            ],
        ),
    ] = Field(default_factory=list)
    """IDs of other bioimage.io resources"""

    uploader: Optional[Uploader] = None
    """The person who uploaded the model (e.g. to bioimage.io)"""

    maintainers: List[Maintainer] = Field(default_factory=list)
    """Maintainers of this resource.
    If not specified, `authors` are maintainers and at least some of them has to specify their `github_user` name"""

    @partial(as_warning, severity=ALERT)
    @field_validator("maintainers", mode="after")
    @classmethod
    def check_maintainers_exist(
        cls, maintainers: List[Maintainer], info: ValidationInfo
    ) -> List[Maintainer]:
        if not maintainers and "authors" in info.data:
            authors: List[Author] = info.data["authors"]
            if all(a.github_user is None for a in authors):
                raise ValueError(
                    "Missing `maintainers` or any author in `authors` with a specified"
                    + " `github_user` name."
                )

        return maintainers

    tags: Annotated[
        List[str],
        Field(examples=[("unet2d", "pytorch", "nucleus", "segmentation", "dsb2018")]),
    ] = Field(default_factory=list)
    """Associated tags"""

    @as_warning
    @field_validator("tags")
    @classmethod
    def warn_about_tag_categories(
        cls, value: List[str], info: ValidationInfo
    ) -> List[str]:
        categories = TAG_CATEGORIES.get(info.data["type"], {})
        missing_categories: List[Dict[str, Sequence[str]]] = []
        for cat, entries in categories.items():
            if not any(e in value for e in entries):
                missing_categories.append({cat: entries})

        if missing_categories:
            raise ValueError(
                f"Missing tags from bioimage.io categories: {missing_categories}"
            )

        return value

    version: Optional[Version] = None
    """The version of the resource following SemVer 2.0."""

    version_number: Optional[int] = None
    """version number (n-th published version, not the semantic version)"""


class GenericDescrBase(GenericModelDescrBase):
    """Base for all resource descriptions except for the model descriptions"""

    format_version: Literal["0.3.0"] = "0.3.0"
    """The **format** version of this resource specification"""

    @model_validator(mode="before")
    @classmethod
    def _convert_from_older_format(
        cls, data: BioimageioYamlContent, /
    ) -> BioimageioYamlContent:
        convert_from_older_format(data)
        return data

    documentation: Annotated[
        Optional[DocumentationSource],
        Field(
            examples=[
                "https://raw.githubusercontent.com/bioimage-io/spec-bioimage-io/main/example_descriptions/models/unet2d_nuclei_broad/README.md",
                "README.md",
            ],
        ),
    ] = None
    """∈📦 URL or relative path to a markdown file with additional documentation.
    The recommended documentation file name is `README.md`. An `.md` suffix is mandatory."""

    badges: List[BadgeDescr] = Field(default_factory=list)
    """badges associated with this resource"""


ResourceDescrType = TypeVar("ResourceDescrType", bound=GenericDescrBase)


class GenericDescr(
    GenericDescrBase, extra="ignore", title="bioimage.io generic specification"
):
    """Specification of the fields used in a generic bioimage.io-compliant resource description file (RDF).

    An RDF is a YAML file that describes a resource such as a model, a dataset, or a notebook.
    Note that those resources are described with a type-specific RDF.
    Use this generic resource description, if none of the known specific types matches your resource.
    """

    type: Annotated[str, LowerCase] = Field("generic", frozen=True)
    """The resource type assigns a broad category to the resource."""

    id: Optional[ResourceId] = None
    """Model zoo (bioimage.io) wide, unique identifier (assigned by bioimage.io)"""

    parent: Optional[ResourceId] = None
    """The description from which this one is derived"""

    source: Optional[HttpUrl] = None
    """The primary source of the resource"""

    @field_validator("type", mode="after")
    @classmethod
    def check_specific_types(cls, value: str) -> str:
        if value in KNOWN_SPECIFIC_RESOURCE_TYPES:
            raise ValueError(
                f"Use the {value} description instead of this generic description for"
                + f" your '{value}' resource."
            )

        return value



# autogen: start
"""
implementaions of all released minor versions are available in submodules:
- notebook v0_2: `bioimageio.spec.notebook.v0_2.NotebookDescr` [user documentation](../../../user_docs/notebook_descr_v0-2.md)
- notebook v0_3: `bioimageio.spec.notebook.v0_3.NotebookDescr` [user documentation](../../../user_docs/notebook_descr_v0-3.md)
"""
from typing import Union

from pydantic import Discriminator
from typing_extensions import Annotated

from .v0_2 import NotebookDescr as NotebookDescr_v0_2
from .v0_3 import NotebookDescr as NotebookDescr
from .v0_3 import NotebookDescr as NotebookDescr_v0_3

AnyNotebookDescr = Annotated[
    Union[NotebookDescr_v0_2, NotebookDescr_v0_3], Discriminator("format_version")
]
"""Union of any released notebook desription"""
# autogen: stop



from typing import Literal, TextIO, Union, cast

from loguru import logger
from pydantic import FilePath, NewPath

from ._description import (
    DISCOVER,
    InvalidDescr,
    ResourceDescr,
    build_description,
    dump_description,
)
from ._internal.common_nodes import ResourceDescrBase
from ._internal.io import BioimageioYamlContent, YamlValue
from ._internal.io_utils import open_bioimageio_yaml, write_yaml
from ._internal.validation_context import ValidationContext
from .common import PermissiveFileSource
from .summary import ValidationSummary


def load_description(
    source: PermissiveFileSource,
    /,
    *,
    format_version: Union[Literal["discover"], Literal["latest"], str] = DISCOVER,
) -> Union[ResourceDescr, InvalidDescr]:
    if isinstance(source, ResourceDescrBase):
        name = getattr(source, "name", f"{str(source)[:10]}...")
        logger.warning("returning already loaded description '{}' as is", name)
        return source  # pyright: ignore[reportReturnType]

    opened = open_bioimageio_yaml(source)

    return build_description(
        opened.content,
        context=ValidationContext(
            root=opened.original_root, file_name=opened.original_file_name
        ),
        format_version=format_version,
    )


def save_bioimageio_yaml_only(
    rd: Union[ResourceDescr, BioimageioYamlContent, InvalidDescr],
    /,
    file: Union[NewPath, FilePath, TextIO],
):
    if isinstance(rd, ResourceDescrBase):
        content = dump_description(rd)
    else:
        content = rd

    write_yaml(cast(YamlValue, content), file)


def load_description_and_validate_format_only(
    source: PermissiveFileSource,
    /,
    *,
    format_version: Union[Literal["discover"], Literal["latest"], str] = DISCOVER,
) -> ValidationSummary:
    rd = load_description(source, format_version=format_version)
    assert rd.validation_summary is not None
    return rd.validation_summary



from itertools import chain
from pathlib import Path
from types import MappingProxyType
from typing import Any, Iterable, List, Literal, Mapping, Tuple, Union, no_type_check

from pydantic import BaseModel, Field, model_validator
from pydantic_core.core_schema import ErrorType
from typing_extensions import TypedDict, assert_never

from ._internal.constants import VERSION
from ._internal.warning_levels import (
    ALERT,
    ALERT_NAME,
    ERROR,
    ERROR_NAME,
    INFO,
    INFO_NAME,
    WARNING,
    WARNING_NAME,
    WarningLevel,
    WarningSeverity,
)

Loc = Tuple[Union[int, str], ...]
"""location of error/warning in a nested data structure"""

WarningSeverityName = Literal["info", "warning", "alert"]
WarningLevelName = Literal[WarningSeverityName, "error"]

WARNING_SEVERITY_TO_NAME: Mapping[WarningSeverity, WarningSeverityName] = (
    MappingProxyType({INFO: INFO_NAME, WARNING: WARNING_NAME, ALERT: ALERT_NAME})
)
WARNING_LEVEL_TO_NAME: Mapping[WarningLevel, WarningLevelName] = MappingProxyType(
    {**WARNING_SEVERITY_TO_NAME, ERROR: ERROR_NAME}
)
WARNING_NAME_TO_LEVEL: Mapping[WarningLevelName, WarningLevel] = MappingProxyType(
    {v: k for k, v in WARNING_LEVEL_TO_NAME.items()}
)


class ValidationEntry(BaseModel):
    loc: Loc
    msg: str
    type: Union[ErrorType, str]


class ErrorEntry(ValidationEntry):
    traceback: List[str] = Field(default_factory=list)


class WarningEntry(ValidationEntry):
    severity: WarningSeverity = WARNING
    severity_name: WarningSeverityName = WARNING_NAME

    @model_validator(mode="before")
    @classmethod
    def sync_severity_with_severity_name(
        cls, data: Union[Mapping[Any, Any], Any]
    ) -> Any:
        if isinstance(data, dict):
            data = dict(data)
            assert isinstance(data, dict)
            if (
                "severity" in data
                and "severity_name" not in data
                and data["severity"] in WARNING_SEVERITY_TO_NAME
            ):
                data["severity_name"] = WARNING_SEVERITY_TO_NAME[data["severity"]]

            if (
                "severity" in data
                and "severity_name" not in data
                and data["severity"] in WARNING_SEVERITY_TO_NAME
            ):
                data["severity"] = WARNING_NAME_TO_LEVEL[data["severity_name"]]

        return data


def format_loc(loc: Loc) -> str:
    if not loc:
        loc = ("__root__",)

    loc_str = ".".join(f"({x})" if x[0].isupper() else x for x in map(str, loc))

    # additional field validation can make the location information quite convoluted, e.g.
    # `weights.pytorch_state_dict.dependencies.source.function-after[validate_url_ok(), url['http','https']]` Input should be a valid URL, relative URL without a base
    # therefore we remove the `.function-after[validate_url_ok(), url['http','https']]` here
    brief_loc_str, *_ = loc_str.split(".function-after")
    return f"`{brief_loc_str}`"


class InstalledPackage(TypedDict):
    name: str
    version: str


class ValidationDetail(BaseModel, extra="allow"):
    name: str
    status: Literal["passed", "failed"]
    errors: List[ErrorEntry] = Field(default_factory=list)
    warnings: List[WarningEntry] = Field(default_factory=list)

    def __str__(self):
        return f"{self.__class__.__name__}:\n" + self.format()

    @property
    def status_icon(self):
        if self.status == "passed":
            return "✔️"
        else:
            return "❌"

    def format(self, hide_tracebacks: bool = False, root_loc: Loc = ()) -> str:
        indent = "    " if root_loc else ""
        errs_wrns = self._format_errors_and_warnings(
            hide_tracebacks=hide_tracebacks, root_loc=root_loc
        )
        return f"{indent}{self.status_icon} {self.name.strip('.')}: {self.status}{errs_wrns}"

    def _format_errors_and_warnings(self, hide_tracebacks: bool, root_loc: Loc):
        indent = "    " if root_loc else ""
        if hide_tracebacks:
            tbs = [""] * len(self.errors)
        else:
            tbs = [
                ("\n      Traceback:\n      " if e.traceback else "")
                + "\n      ".join(e.traceback)
                for e in self.errors
            ]

        def join_parts(parts: Iterable[Tuple[str, str]]):
            last_loc = None
            lines: List[str] = []
            for loc, msg in parts:
                if loc == last_loc:
                    lines.append(f"\n  {loc} {msg}")
                else:
                    lines.append(f"\n- {loc} {msg}")

                last_loc = loc

            return "".join(lines)

        es = join_parts(
            (format_loc(root_loc + e.loc), f"{e.msg}{tb}")
            for e, tb in zip(self.errors, tbs)
        )
        ws = join_parts((format_loc(root_loc + w.loc), w.msg) for w in self.warnings)

        return (
            f"\n{indent}errors:\n{es}"
            if es
            else "" + f"\n{indent}warnings:\n{ws}" if ws else ""
        )


class ValidationSummary(BaseModel, extra="allow"):
    name: str
    source_name: str
    status: Literal["passed", "failed"]
    details: List[ValidationDetail]
    env: List[InstalledPackage] = Field(
        default_factory=lambda: [
            InstalledPackage(name="bioimageio.spec", version=VERSION)
        ]
    )
    """list of selected, relevant package versions"""

    @property
    def status_icon(self):
        if self.status == "passed":
            return "✔️"
        else:
            return "❌"

    @property
    def errors(self) -> List[ErrorEntry]:
        return list(chain.from_iterable(d.errors for d in self.details))

    @property
    def warnings(self) -> List[WarningEntry]:
        return list(chain.from_iterable(d.warnings for d in self.details))

    def __str__(self):
        return f"{self.__class__.__name__}:\n" + self.format()

    @staticmethod
    def _format_md_table(rows: List[List[str]]) -> str:
        """format `rows` as markdown table"""
        n_cols = len(rows[0])
        assert all(len(row) == n_cols for row in rows)
        col_widths = [max(len(row[i]) for row in rows) for i in range(n_cols)]

        lines = [" | ".join(rows[0][i].center(col_widths[i]) for i in range(n_cols))]
        lines.append(" | ".join("---".center(col_widths[i]) for i in range(n_cols)))
        lines.extend(
            [
                " | ".join(row[i].ljust(col_widths[i]) for i in range(n_cols))
                for row in rows[1:]
            ]
        )
        return "\n| " + " |\n| ".join(lines) + " |\n"

    def _format_env(self):
        if not self.env:
            return ""

        rows = [["package", "version"]] + [[e["name"], e["version"]] for e in self.env]
        return self._format_md_table(rows)

    def format(
        self,
        hide_tracebacks: bool = False,
        hide_source: bool = False,
        hide_env: bool = False,
        root_loc: Loc = (),
    ) -> str:
        indent = "   " if root_loc else ""
        src = "" if hide_source else f"\n{indent}source: {self.source_name}"
        env = "" if hide_env else self._format_env()
        details = [["❓", "location", "detail"]]

        def format_loc(loc: Loc):
            return "`" + (".".join(map(str, root_loc + loc)) or ".") + "`"

        for d in self.details:
            details.append([d.status_icon, "", d.name])
            for entry in d.errors:
                details.append(["❌", format_loc(entry.loc), entry.msg])
                if hide_tracebacks:
                    continue

                for tb in entry.traceback:
                    if not (tb_stripped := tb.strip()):
                        continue

                    first_tb_line, *tb_lines = tb_stripped.split("\n")
                    if (
                        first_tb_line.startswith('File "')
                        and '", line' in first_tb_line
                    ):
                        path, where = first_tb_line[len('File "') :].split('", line')
                        try:
                            p = Path(path)
                        except Exception:
                            file_name = path
                        else:
                            path = p.as_posix()
                            file_name = p.name

                        where = ", line" + where
                        first_tb_line = f'[{file_name}]({file_name} "{path}"){where}'

                    if tb_lines:
                        tb_rest = "<br>`" + "`<br>`".join(tb_lines) + "`"
                    else:
                        tb_rest = ""

                    details.append(["", "", first_tb_line + tb_rest])

            if d.errors:
                details.append(["", "", ""])

            for entry in d.warnings:
                details.append(["⚠", format_loc(entry.loc), entry.msg])
            if d.warnings:
                details.append(["", "", ""])

        return (
            f"{indent}{self.status_icon} {self.name.strip('.')}: {self.status}\n"
            + f"{src}{env}\n{self._format_md_table(details)}"
        )

    @no_type_check
    def display(self) -> None:
        formatted = self.format()
        try:
            from IPython.core.getipython import get_ipython
            from IPython.display import Markdown, display
        except ImportError:
            print(formatted)
        else:
            if get_ipython() is None:
                print(formatted)
            else:
                _ = display(Markdown(formatted))

    def add_detail(self, detail: ValidationDetail):
        if detail.status == "failed":
            self.status = "failed"
        elif detail.status != "passed":
            assert_never(detail.status)

        self.details.append(detail)



from typing import Any, Dict

import pytest
from annotated_types import Ge
from typing_extensions import Annotated

from bioimageio.spec._internal.common_nodes import Node
from bioimageio.spec._internal.field_warning import warn
from bioimageio.spec._internal.validation_context import ValidationContext
from bioimageio.spec._internal.warning_levels import (
    ALERT,
    ERROR,
    INFO,
    WARNING,
)
from tests.utils import WARNING_LEVEL_CONTEXT_KEY, check_node


class DummyNode(Node):
    a: Annotated[int, warn(Ge(0), "smaller than zero")] = 0
    b: Annotated[int, warn(Ge(0), "smaller than zero", WARNING)] = 0


class NestedDummyNode(Node):
    dummy: DummyNode


DUMMY_INPUT: Dict[str, Any] = {"a": -1, "b": -1}
NESTED_DICT_DUMMY_INPUT = dict(dummy=DUMMY_INPUT)
NESTED_NODE_DUMMY_INPUT = dict(dummy=DummyNode(**DUMMY_INPUT))


@pytest.mark.parametrize(
    "kwargs,context,valid",
    [
        (DUMMY_INPUT, {}, True),
        (DUMMY_INPUT, {WARNING_LEVEL_CONTEXT_KEY: ERROR}, True),
        (DUMMY_INPUT, {WARNING_LEVEL_CONTEXT_KEY: ALERT}, True),
        (DUMMY_INPUT, {WARNING_LEVEL_CONTEXT_KEY: WARNING}, False),
        (DUMMY_INPUT, {WARNING_LEVEL_CONTEXT_KEY: INFO}, False),
    ],
)
def test_warn(kwargs: Dict[str, Any], context: ValidationContext, valid: bool):
    check_node(DummyNode, kwargs, context=context, is_invalid=not valid)


@pytest.mark.parametrize(
    "kwargs,context,valid",
    [
        (NESTED_DICT_DUMMY_INPUT, {}, True),
        (NESTED_DICT_DUMMY_INPUT, {WARNING_LEVEL_CONTEXT_KEY: ERROR}, True),
        (NESTED_DICT_DUMMY_INPUT, {WARNING_LEVEL_CONTEXT_KEY: ALERT}, True),
        (NESTED_NODE_DUMMY_INPUT, {}, True),
        (NESTED_NODE_DUMMY_INPUT, {WARNING_LEVEL_CONTEXT_KEY: ERROR}, True),
        (NESTED_NODE_DUMMY_INPUT, {WARNING_LEVEL_CONTEXT_KEY: ALERT}, True),
        (NESTED_DICT_DUMMY_INPUT, {WARNING_LEVEL_CONTEXT_KEY: WARNING}, False),
        (NESTED_DICT_DUMMY_INPUT, {WARNING_LEVEL_CONTEXT_KEY: INFO}, False),
        (
            NESTED_NODE_DUMMY_INPUT,
            {WARNING_LEVEL_CONTEXT_KEY: WARNING},
            True,  # no reevaluation of node instance
        ),
        (
            NESTED_NODE_DUMMY_INPUT,
            {WARNING_LEVEL_CONTEXT_KEY: INFO},
            True,  # no reevaluation of node instance
        ),
    ],
)
def test_warn_nested(kwargs: Dict[str, Any], context: ValidationContext, valid: bool):
    check_node(NestedDummyNode, kwargs, context=context, is_invalid=not valid)



# autogen: start
"""
implementaions of all released minor versions are available in submodules:
- application v0_2: `bioimageio.spec.application.v0_2.ApplicationDescr` [user documentation](../../../user_docs/application_descr_v0-2.md)
- application v0_3: `bioimageio.spec.application.v0_3.ApplicationDescr` [user documentation](../../../user_docs/application_descr_v0-3.md)
"""
from typing import Union

from pydantic import Discriminator
from typing_extensions import Annotated

from .v0_2 import ApplicationDescr as ApplicationDescr_v0_2
from .v0_3 import ApplicationDescr as ApplicationDescr
from .v0_3 import ApplicationDescr as ApplicationDescr_v0_3

AnyApplicationDescr = Annotated[
    Union[ApplicationDescr_v0_2, ApplicationDescr_v0_3], Discriminator("format_version")
]
"""Union of any released application desription"""
# autogen: stop



from ._internal.io import download as download
from ._internal.io import (
    identify_bioimageio_yaml_file_name as identify_bioimageio_yaml_file_name,
)
from ._internal.io import is_valid_bioimageio_yaml_name as is_valid_bioimageio_yaml_name
from ._internal.io_utils import load_array as load_array
from ._internal.io_utils import save_array as save_array



"""script that updates the content of bioimageio/spec/static/spdx_licenses.json and generates the `LicenseId` Literal"""

import json
import sys
import urllib.request
from argparse import ArgumentParser
from pathlib import Path

import black.files
import black.mode

PROJECT_ROOT = Path(__file__).parent.parent

URL = (
    "https://raw.githubusercontent.com/spdx/license-list-data/{tag}/json/licenses.json"
)
LICENSES_JSON_FILE = PROJECT_ROOT / "bioimageio/spec/static/spdx_licenses.json"
LICENSE_ID_MODULE_PATH = (
    PROJECT_ROOT / "bioimageio/spec/_internal/_generated_spdx_license_literals.py"
)
LICENSE_ID_MODULE_TEMPLATE = """# This file was generated by scripts/update_spdx_licenses.py
from typing import Literal

LicenseId = Literal{license_ids}

DeprecatedLicenseId = Literal{deprecated_license_ids}
"""


def parse_args():
    p = ArgumentParser(description="script that generates weights formats overview")
    _ = p.add_argument("tag", nargs="?", default="v3.21")

    args = p.parse_args()
    return dict(tag=args.tag)


def main(*, tag: str):
    url = URL.format(tag=tag)
    print("requesting:", url)
    text = urllib.request.urlopen(url).read().decode("utf-8")
    _ = LICENSES_JSON_FILE.write_text(text, encoding="utf-8")
    print(f"Updated {LICENSES_JSON_FILE}")

    licenses = json.loads(text)["licenses"]
    license_ids = [x["licenseId"] for x in licenses if not x["isDeprecatedLicenseId"]]
    deprecated_license_ids = [
        x["licenseId"] for x in licenses if x["isDeprecatedLicenseId"]
    ]
    code = LICENSE_ID_MODULE_TEMPLATE.format(
        license_ids=license_ids, deprecated_license_ids=deprecated_license_ids
    )

    # apply black formating
    black_config = black.files.parse_pyproject_toml(
        str(PROJECT_ROOT / "pyproject.toml")
    )
    black_config["target_versions"] = set(
        (
            getattr(black.mode.TargetVersion, tv.upper())
            for tv in black_config.pop("target_version")
        )
    )
    code = black.format_str(code, mode=black.mode.Mode(**black_config))

    _ = LICENSE_ID_MODULE_PATH.write_text(code, encoding="utf-8")
    print(f"Updated {LICENSE_ID_MODULE_PATH}")


if __name__ == "__main__":
    kwargs = parse_args()
    sys.exit(main(**kwargs))



from pathlib import Path

import pytest
from pydantic import TypeAdapter, ValidationError
from typing_extensions import Annotated

from bioimageio.spec._internal.io import WithSuffix
from bioimageio.spec._internal.types import FileSource
from bioimageio.spec._internal.validation_context import ValidationContext


def test_single_suffix():
    adapter = TypeAdapter(Annotated[FileSource, WithSuffix(".py", case_sensitive=True)])
    with ValidationContext(root=Path(__file__).parent):
        _ = adapter.validate_python(Path(__file__).name)

    with ValidationContext(perform_io_checks=False):
        _ = adapter.validate_python("https://example.com/lala.py")
        _ = adapter.validate_python("https://example.com/lala.py#section")


def test_case_sensitive_suffix():
    adapter = TypeAdapter(Annotated[FileSource, WithSuffix(".py", case_sensitive=True)])
    with ValidationContext(perform_io_checks=False), pytest.raises(ValidationError):
        _ = adapter.validate_python("https://example.com/lala.PY")


def test_multiple_suffix():
    adapter = TypeAdapter(
        Annotated[FileSource, WithSuffix((".py", ".md"), case_sensitive=True)]
    )
    with ValidationContext(root=Path(__file__).parent):
        _ = adapter.validate_python(Path(__file__).name)

    with ValidationContext(perform_io_checks=False):
        _ = adapter.validate_python("https://example.com/lala.py")
        _ = adapter.validate_python("https://example.com/lala.md#section")



from functools import partial
from types import MappingProxyType
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Literal,
    Optional,
    Union,
    cast,
    get_args,
)

from pydantic import PrivateAttr, model_validator
from typing_extensions import Self

from .._build_description import build_description_impl, get_rd_class_impl
from .._internal.common_nodes import InvalidDescr, Node
from .._internal.field_warning import issue_warning
from .._internal.io import FileDescr as FileDescr
from .._internal.io import Sha256 as Sha256
from .._internal.io import YamlValue
from .._internal.io_basics import AbsoluteFilePath as AbsoluteFilePath
from .._internal.io_utils import open_bioimageio_yaml
from .._internal.types import FileSource, NotEmpty
from .._internal.url import HttpUrl as HttpUrl
from .._internal.validation_context import (
    validation_context_var,
)
from .._internal.warning_levels import ALERT
from ..application import ApplicationDescr_v0_2, ApplicationDescr_v0_3
from ..dataset import DatasetDescr_v0_2, DatasetDescr_v0_3
from ..generic import GenericDescr_v0_2, GenericDescr_v0_3
from ..generic.v0_3 import VALID_COVER_IMAGE_EXTENSIONS as VALID_COVER_IMAGE_EXTENSIONS
from ..generic.v0_3 import Author as Author
from ..generic.v0_3 import BadgeDescr as BadgeDescr
from ..generic.v0_3 import CiteEntry as CiteEntry
from ..generic.v0_3 import Doi as Doi
from ..generic.v0_3 import (
    GenericDescrBase,
    ResourceId,
    _author_conv,  # pyright: ignore[reportPrivateUsage]
    _maintainer_conv,  # pyright: ignore[reportPrivateUsage]
)
from ..generic.v0_3 import LinkedResource as LinkedResource
from ..generic.v0_3 import Maintainer as Maintainer
from ..generic.v0_3 import OrcidId as OrcidId
from ..generic.v0_3 import RelativeFilePath as RelativeFilePath
from ..generic.v0_3 import Uploader as Uploader
from ..generic.v0_3 import Version as Version
from ..model import ModelDescr_v0_4, ModelDescr_v0_5
from ..notebook import NotebookDescr_v0_2, NotebookDescr_v0_3
from .v0_2 import CollectionDescr as _CollectionDescr_v0_2


class CollectionId(ResourceId):
    pass


EntryDescr = Union[
    ApplicationDescr_v0_2,
    ApplicationDescr_v0_3,
    DatasetDescr_v0_2,
    DatasetDescr_v0_3,
    ModelDescr_v0_4,
    ModelDescr_v0_5,
    NotebookDescr_v0_2,
    NotebookDescr_v0_3,
    GenericDescr_v0_2,
    GenericDescr_v0_3,
]

_ENTRY_DESCR_MAP = MappingProxyType(
    {
        None: MappingProxyType(
            {
                "0.2": GenericDescr_v0_2,
                "0.3": GenericDescr_v0_3,
                None: GenericDescr_v0_3,
            }
        ),
        "generic": MappingProxyType(
            {
                "0.2": GenericDescr_v0_2,
                "0.3": GenericDescr_v0_3,
                None: GenericDescr_v0_3,
            }
        ),
        "application": MappingProxyType(
            {
                "0.2": ApplicationDescr_v0_2,
                "0.3": ApplicationDescr_v0_3,
                None: ApplicationDescr_v0_3,
            }
        ),
        "dataset": MappingProxyType(
            {
                "0.2": DatasetDescr_v0_2,
                "0.3": DatasetDescr_v0_3,
                None: DatasetDescr_v0_3,
            }
        ),
        "notebook": MappingProxyType(
            {
                "0.2": NotebookDescr_v0_2,
                "0.3": NotebookDescr_v0_3,
                None: NotebookDescr_v0_3,
            }
        ),
        "model": MappingProxyType(
            {
                "0.3": ModelDescr_v0_4,
                "0.4": ModelDescr_v0_4,
                "0.5": ModelDescr_v0_5,
                None: ModelDescr_v0_5,
            }
        ),
    }
)


class CollectionEntry(Node, extra="allow"):
    """A collection entry description is based on the collection description itself.
    Fields are added/overwritten by the content of `descr_source` if `descr_source` is set,
    and finally added/overwritten by any fields specified directly in the entry.
    Except for the `id` field, fields are overwritten entirely, their content is not merged!
    The final `id` for each collection entry is composed of the collection's `id`
    and the entry's 'sub-'`id`, specified externally in `descr_source` or superseeded in-place,
    such that the `final_entry_id = <collection_id>/<entry_id>`"""

    entry_source: Optional[FileSource] = None
    """an external source this entry description is based on"""

    id: Optional[ResourceId] = None
    """Collection entry sub id overwriting `rdf_source.id`.
    The full collection entry's id is the collection's base id, followed by this sub id and separated by a slash '/'."""

    _descr: Optional[EntryDescr] = PrivateAttr(None)

    @property
    def entry_update(self) -> Dict[str, YamlValue]:
        return self.model_extra or {}

    @property
    def descr(self) -> Optional[EntryDescr]:
        if self._descr is None:
            issue_warning(
                "Collection entry description not set. Is this entry part of a"
                + " Collection? A collection entry only has its `descr` set if it is part"
                + " of a valid collection description.",
                value=None,
                severity=ALERT,
            )

        return self._descr


class CollectionDescr(
    GenericDescrBase, extra="allow", title="bioimage.io collection specification"
):
    """A bioimage.io collection resource description file (collection RDF) describes a collection of bioimage.io
    resources.
    The resources listed in a collection RDF have types other than 'collection'; collections cannot be nested.
    """

    type: Literal["collection"] = "collection"

    id: Optional[CollectionId] = None
    """Model zoo (bioimage.io) wide, unique identifier (assigned by bioimage.io)"""

    parent: Optional[CollectionId] = None
    """The description from which this one is derived"""

    collection: NotEmpty[List[CollectionEntry]]
    """Collection entries"""

    @model_validator(mode="after")
    def finalize_entries(self) -> Self:
        context = validation_context_var.get()
        common_entry_content = {
            k: v
            for k, v in self.model_dump(mode="json", exclude_unset=True).items()
            if k not in ("id", "collection")
        }
        common_badges = common_entry_content.pop(
            "badges", None
        )  # `badges` not valid for model entries
        base_id: Optional[CollectionId] = self.id

        seen_entry_ids: Dict[str, int] = {}

        for i, entry in enumerate(self.collection):
            entry_data: Dict[str, Any] = dict(common_entry_content)
            # set entry specific root as it might be adapted in the presence of an external entry source
            entry_root = context.root
            entry_file_name = context.file_name

            if entry.entry_source is not None:
                if not context.perform_io_checks:
                    issue_warning(
                        "Skipping IO relying validation for collection[{i}]",
                        value=entry.entry_source,
                        msg_context=dict(i=i),
                    )
                    continue

                external_data = open_bioimageio_yaml(entry.entry_source)
                # add/overwrite common collection entry content with external source
                entry_data.update(external_data.content)
                entry_root = external_data.original_root
                entry_file_name = external_data.original_file_name

            # add/overwrite common+external entry content with in-place entry update
            entry_data.update(entry.entry_update)

            # also update explicitly specified `id` field data
            if entry.id is not None:
                entry_data["id"] = entry.id

            if "id" in entry_data:
                entry_id = str(entry_data["id"])
                if (seen_i := seen_entry_ids.get(entry_id)) is not None:
                    raise ValueError(
                        f"Dublicate `id` '{entry_data['id']}' in"
                        + f" collection[{seen_i}]/collection[{i}]"
                    )

                seen_entry_ids[entry_id] = i
            else:
                raise ValueError(f"Missing `id` for entry {i}")

            if base_id is not None:
                entry_data["id"] = f"{base_id}/{entry_data['id']}"

            type_ = entry_data.get("type")
            if type_ == "collection":
                raise ValueError(
                    f"collection[{i}].type may not be 'collection'; collections may not"
                    + " be nested!"
                )

            if (
                type_ != "model"
                and common_badges is not None
                and "badges" not in entry_data
            ):
                # set badges from the collection root for non-model resources if not set for this specific entry
                entry_data["badges"] = common_badges

            entry_descr = build_description_impl(
                entry_data,
                context=context.replace(root=entry_root, file_name=entry_file_name),
                get_rd_class=partial(
                    get_rd_class_impl, descriptions_map=_ENTRY_DESCR_MAP
                ),
            )

            assert entry_descr.validation_summary is not None
            if isinstance(entry_descr, InvalidDescr):
                raise ValueError(
                    "Invalid collection entry"
                    + f" collection[{i}]:\n"
                    + f"{entry_descr.validation_summary.format(hide_source=True, hide_env=True, root_loc=('collection', i))}"
                )
            elif isinstance(
                entry_descr, get_args(EntryDescr)
            ):  # TODO: use EntryDescr as union (py>=3.10)
                entry._descr = entry_descr  # type: ignore
            else:
                raise ValueError(
                    f"{entry_descr.type} {entry_descr.format_version} entries "
                    + f"are not allowed in {self.type} {self.format_version}."
                )
        return self

    @model_validator(mode="before")
    @classmethod
    def _convert(cls, data: Dict[str, Any], /) -> Dict[str, Any]:
        if (
            data.get("type") == "collection"
            and isinstance(fv := data.get("format_version"), str)
            and fv.startswith("0.2.")
        ):
            old = _CollectionDescr_v0_2.load(data)
            if isinstance(old, InvalidDescr):
                return data

            return cast(
                Dict[str, Any],
                (cls if TYPE_CHECKING else dict)(
                    attachments=(
                        []
                        if old.attachments is None
                        else [FileDescr(source=f) for f in old.attachments.files]
                    ),
                    authors=[_author_conv.convert_as_dict(a) for a in old.authors],
                    badges=old.badges,
                    cite=[
                        {"text": c.text, "doi": c.doi, "url": c.url} for c in old.cite
                    ],
                    collection=[
                        (CollectionEntry if TYPE_CHECKING else dict)(
                            entry_source=entry.rdf_source, id=entry.id, **entry.rdf_update  # type: ignore
                        )
                        for entry in old.collection
                    ],
                    config=old.config,
                    covers=old.covers,
                    description=old.description,
                    documentation=old.documentation,
                    format_version="0.3.0",
                    git_repo=old.git_repo,
                    icon=old.icon,
                    id=old.id,
                    license=old.license,
                    links=old.links,
                    maintainers=[
                        _maintainer_conv.convert_as_dict(m) for m in old.maintainers
                    ],
                    name=old.name,
                    tags=old.tags,
                    type=old.type,
                    uploader=old.uploader,
                    version=old.version,
                    **(old.model_extra or {}),
                ),
            )

        return data


class LinkedCollection(Node):
    """Reference to a bioimage.io collection."""

    id: CollectionId
    """A valid collection `id` from the bioimage.io collection."""

    version_number: int
    """version number (n-th published version, not the semantic version) of linked collection"""



from pathlib import Path

from annotated_types import Predicate
from pydantic import DirectoryPath, FilePath
from typing_extensions import Annotated

FileName = str
AbsoluteDirectory = Annotated[DirectoryPath, Predicate(Path.is_absolute)]
AbsoluteFilePath = Annotated[FilePath, Predicate(Path.is_absolute)]

BIOIMAGEIO_YAML = "bioimageio.yaml"
ALTERNATIVE_BIOIMAGEIO_YAML_NAMES = ("rdf.yaml", "model.yaml")
ALL_BIOIMAGEIO_YAML_NAMES = (BIOIMAGEIO_YAML,) + ALTERNATIVE_BIOIMAGEIO_YAML_NAMES



import io
import warnings
from contextlib import nullcontext
from pathlib import Path
from typing import IO, Any, Dict, Mapping, Optional, TextIO, Union, cast
from zipfile import ZipFile, is_zipfile

import numpy
from numpy.typing import NDArray
from pydantic import DirectoryPath, FilePath, NewPath
from ruyaml import YAML
from typing_extensions import Unpack

from .io import (
    BIOIMAGEIO_YAML,
    BioimageioYamlContent,
    FileDescr,
    HashKwargs,
    OpenedBioimageioYaml,
    YamlValue,
    download,
    find_bioimageio_yaml_file_name,
)
from .io_basics import FileName
from .types import FileSource, PermissiveFileSource

yaml = YAML(typ="safe")


def read_yaml(file: Union[FilePath, TextIO]) -> YamlValue:
    if isinstance(file, Path):
        cm = file.open("r", encoding="utf-8")
    else:
        cm = nullcontext(file)

    with cm as f:
        content: YamlValue = yaml.load(f)

    return content


def write_yaml(content: YamlValue, /, file: Union[NewPath, FilePath, TextIO]):
    if isinstance(file, Path):
        cm = file.open("w", encoding="utf-8")
    else:
        cm = nullcontext(file)

    with cm as f:
        yaml.dump(content, f)


def _sanitize_bioimageio_yaml(content: YamlValue) -> BioimageioYamlContent:
    if not isinstance(content, dict):
        raise ValueError(
            f"Expected {BIOIMAGEIO_YAML} content to be a mapping (got {type(content)})."
        )

    for key in content:
        if not isinstance(key, str):
            raise ValueError(
                f"Expected all keys (field names) in a {BIOIMAGEIO_YAML} "
                + f"need to be strings (got '{key}' of type {type(key)})."
            )

    return cast(BioimageioYamlContent, content)


def open_bioimageio_yaml(
    source: PermissiveFileSource, /, **kwargs: Unpack[HashKwargs]
) -> OpenedBioimageioYaml:
    downloaded = download(source, **kwargs)
    local_source = downloaded.path
    root = downloaded.original_root

    if is_zipfile(local_source):
        local_source = unzip(local_source)

    if local_source.is_dir():
        root = local_source
        local_source = local_source / find_bioimageio_yaml_file_name(local_source)

    content = _sanitize_bioimageio_yaml(read_yaml(local_source))

    return OpenedBioimageioYaml(content, root, downloaded.original_file_name)


def unzip(
    zip_file: Union[FilePath, ZipFile],
    out_path: Optional[DirectoryPath] = None,
    overwrite: bool = False,
) -> DirectoryPath:
    if isinstance(zip_file, ZipFile):
        zip_context = nullcontext(zip_file)
        if out_path is None:
            raise ValueError("Missing argument: out_path")
    else:
        zip_context = ZipFile(zip_file, "r")
        if out_path is None:
            out_path = zip_file.with_suffix(zip_file.suffix + ".unzip")

    with zip_context as f:
        if out_path.exists() and overwrite:
            if overwrite:
                warnings.warn(f"Overwriting existing unzipped archive at {out_path}")
            else:
                found_content = {
                    p.relative_to(out_path).as_posix() for p in out_path.glob("*")
                }
                expected_content = {info.filename for info in f.filelist}
                if expected_content - found_content:
                    warnings.warn(
                        f"Unzipped archive at {out_path} is missing expected files."
                    )
                    parts = out_path.name.split("_")
                    nr, *suffixes = parts[-1].split(".")
                    if nr.isdecimal():
                        nr = str(int(nr) + 1)
                    else:
                        nr = f"1.{nr}"

                    parts[-1] = ".".join([nr, *suffixes])
                    return unzip(
                        f, out_path.with_name("_".join(parts)), overwrite=overwrite
                    )
                else:
                    warnings.warn(
                        "Using already unzipped archive with all expected files at"
                        + f" {out_path}."
                    )
                    return out_path

        f.extractall(out_path)

    return out_path


def write_zip(
    path: Union[FilePath, IO[bytes]],
    content: Mapping[FileName, Union[str, FilePath, Dict[Any, Any]]],
    *,
    compression: int,
    compression_level: int,
) -> None:
    """Write a zip archive.

    Args:
        path: output path to write to.
        content: dict mapping archive names to local file paths, strings (for text files), or dict (for yaml files).
        compression: The numeric constant of compression method.
        compression_level: Compression level to use when writing files to the archive.
                           See https://docs.python.org/3/library/zipfile.html#zipfile.ZipFile

    """
    with ZipFile(
        path, "w", compression=compression, compresslevel=compression_level
    ) as myzip:
        for arc_name, file in content.items():
            if isinstance(file, dict):
                buf = io.StringIO()
                write_yaml(file, buf)
                file = buf.getvalue()

            if isinstance(file, str):
                myzip.writestr(arc_name, file.encode("utf-8"))
            else:
                myzip.write(file, arcname=arc_name)


def load_array(source: Union[FileSource, FileDescr]) -> NDArray[Any]:
    path = download(source).path

    return numpy.load(path, allow_pickle=False)


def save_array(path: Path, array: NDArray[Any]) -> None:
    return numpy.save(path, array, allow_pickle=False)



import re
import sys
from argparse import ArgumentParser
from dataclasses import dataclass, field
from difflib import ndiff
from pathlib import Path
from typing import List, Literal

import black.files
import black.mode

ROOT_PATH = Path(__file__).parent.parent

AUTOGEN_START = "# autogen: start\n"
AUTOGEN_BODY_SINGLE = """from . import {info.all_version_modules_import_as}
from .{info.latest_version_module} import {info.target_node} as {info.target_node}

Any{info.target_node} = {info.target_node}
"""
AUTOGEN_BODY_MULTIPLE = """\"\"\"
implementaions of all released minor versions are available in submodules:
{info.submodule_list}
\"\"\"
from typing import Union

from pydantic import Discriminator
from typing_extensions import Annotated

{info.all_version_modules_imports}

Any{info.target_node} = Annotated[Union[{info.all_target_nodes_plain_aliases}], Discriminator("format_version")]
\"\"\"Union of any released {info.target} desription\"\"\"
"""

AUTOGEN_STOP = "# autogen: stop\n"

VERSION_MODULE_PATTERN = r"v(?P<major>\d+)_(?P<minor>\d+).py"


def main(command: Literal["check", "generate"]):
    for target in [
        "generic",
        "model",
        "dataset",
        "collection",
        "notebook",
        "application",
    ]:
        process(
            Info(
                target=target,
                all_version_modules=get_ordered_version_submodules(target),
            ),
            check=command == "check",
        )

    return 0


def parse_args():
    p = ArgumentParser(
        description=(
            "script that generates imports in bioimageio.spec resource description"
            + " submodules"
        )
    )
    _ = p.add_argument(
        "command", choices=["check", "generate"], nargs="?", default="generate"
    )
    args = p.parse_args()
    return args


@dataclass
class Info:
    target: str
    all_version_modules: List[str]
    target_node: str = field(init=False)
    all_target_nodes_plain: str = field(init=False)
    all_target_nodes_plain_aliases: str = field(init=False)
    latest_version_module: str = field(init=False)
    all_version_modules_import_as: str = field(init=False)
    all_version_modules_imports: str = field(init=False)
    package_path: Path = field(init=False)
    submodule_list: str = field(init=False)

    def __post_init__(self):
        self.target_node = self.target.capitalize() + "Descr"
        self.all_target_nodes_plain = ", ".join(
            [f"{vm}.{self.target_node}" for vm in self.all_version_modules]
        )
        self.all_target_nodes_plain_aliases = ", ".join(
            [f"{self.target_node}_{vm}" for vm in self.all_version_modules]
        )
        self.latest_version_module = self.all_version_modules[-1]
        self.all_version_modules_import_as = ", ".join(
            f"{m} as {m}" for m in self.all_version_modules
        )

        avmi = [
            f"from .{m} import {self.target_node} as {self.target_node}_{m}"
            for m in self.all_version_modules
        ]
        avmi.insert(
            -1,
            f"from .{self.latest_version_module} import {self.target_node} as "
            + f"{self.target_node}",
        )
        self.all_version_modules_imports = "\n".join(avmi)

        self.package_path = (ROOT_PATH / "bioimageio" / "spec" / self.target).resolve()
        self.submodule_list = "\n".join(
            [
                f"- {self.target} {vm}: `bioimageio.spec.{self.target}.{vm}."
                + f"{self.target_node}` [user documentation](../../../user_docs/"
                + f"{self.target}_descr_{vm.replace('_', '-')}.md)"
                for vm in self.all_version_modules
            ]
        )


def process(info: Info, check: bool):
    package_init = info.package_path / "__init__.py"
    print(f"{'Checking' if check else 'Updating'} {package_init}")

    init_content = package_init.read_text()
    pattern = AUTOGEN_START + ".*" + AUTOGEN_STOP
    flags = re.DOTALL
    if not re.findall(pattern, init_content, flags=flags):
        raise RuntimeError(
            f"Could not find autogen markers in {package_init}. Expected to"
            + f" find:\n{AUTOGEN_START}...{AUTOGEN_STOP}"
        )

    autogen_body = (
        AUTOGEN_BODY_SINGLE
        if len(info.all_version_modules) == 1
        else AUTOGEN_BODY_MULTIPLE
    )
    updated = re.sub(
        pattern,
        AUTOGEN_START + autogen_body.format(info=info) + AUTOGEN_STOP,
        init_content,
        flags=flags,
    )
    black_config = black.files.parse_pyproject_toml(str(ROOT_PATH / "pyproject.toml"))
    black_config["target_versions"] = set(
        (
            getattr(black.mode.TargetVersion, tv.upper())
            for tv in black_config.pop("target_version")
        )
    )
    updated = black.format_str(updated, mode=black.mode.Mode(**black_config))
    if check:
        if init_content == updated:
            print("all seems fine")
        else:
            raise RuntimeError(
                "call with mode 'generate' to update:\n"
                + "".join(
                    ndiff(
                        init_content.splitlines(keepends=True),
                        updated.splitlines(keepends=True),
                    )
                )
            )
    else:
        with package_init.open("w", newline="\n") as f:
            _ = f.write(updated)


def get_ordered_version_submodules(target: str):
    matches = [
        m
        for p in (ROOT_PATH / "bioimageio" / "spec" / target).iterdir()
        if p.is_file() and (m := re.fullmatch(VERSION_MODULE_PATTERN, p.name))
    ]
    if not matches:
        raise RuntimeError(f"No version modules found for target '{target}'")

    return [
        m.string[:-3]
        for m in sorted(matches, key=lambda m: (int(m["major"]), int(m["minor"])))
    ]


if __name__ == "__main__":
    args = parse_args()
    sys.exit(main(args.command))



# autogen: start
"""
implementaions of all released minor versions are available in submodules:
- dataset v0_2: `bioimageio.spec.dataset.v0_2.DatasetDescr` [user documentation](../../../user_docs/dataset_descr_v0-2.md)
- dataset v0_3: `bioimageio.spec.dataset.v0_3.DatasetDescr` [user documentation](../../../user_docs/dataset_descr_v0-3.md)
"""
from typing import Union

from pydantic import Discriminator
from typing_extensions import Annotated

from .v0_2 import DatasetDescr as DatasetDescr_v0_2
from .v0_3 import DatasetDescr as DatasetDescr
from .v0_3 import DatasetDescr as DatasetDescr_v0_3

AnyDatasetDescr = Annotated[
    Union[DatasetDescr_v0_2, DatasetDescr_v0_3], Discriminator("format_version")
]
"""Union of any released dataset desription"""
# autogen: stop



from __future__ import annotations

import hashlib
import sys
import warnings
from abc import abstractmethod
from dataclasses import dataclass
from datetime import date as _date
from datetime import datetime as _datetime
from pathlib import Path, PurePath
from typing import (
    Any,
    ClassVar,
    Dict,
    Generic,
    Iterable,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypedDict,
    TypeVar,
    Union,
)
from urllib.parse import urlparse, urlsplit, urlunsplit
from zipfile import ZipFile, is_zipfile

import pooch
import pydantic
from pydantic import (
    AnyUrl,
    DirectoryPath,
    FilePath,
    GetCoreSchemaHandler,
    PlainSerializer,
    PrivateAttr,
    RootModel,
    SerializationInfo,
    StringConstraints,
    TypeAdapter,
    model_validator,
)
from pydantic_core import core_schema
from typing_extensions import (
    Annotated,
    LiteralString,
    NotRequired,
    Self,
    Unpack,
    assert_never,
)
from typing_extensions import TypeAliasType as _TypeAliasType

from .._internal._settings import settings
from .._internal.io_basics import (
    ALL_BIOIMAGEIO_YAML_NAMES,
    ALTERNATIVE_BIOIMAGEIO_YAML_NAMES,
    BIOIMAGEIO_YAML,
    AbsoluteDirectory,
    AbsoluteFilePath,
    FileName,
)
from .._internal.node import Node
from .._internal.packaging_context import packaging_context_var
from .._internal.root_url import RootHttpUrl
from .._internal.url import HttpUrl
from .._internal.validated_string import ValidatedString
from .._internal.validation_context import (
    validation_context_var,
)
from .validator_annotations import AfterValidator

if sys.version_info < (3, 10):
    SLOTS: Dict[str, bool] = {}
else:
    SLOTS = {"slots": True}


class Sha256(ValidatedString):
    """SHA-256 hash value"""

    root_model: ClassVar[Type[RootModel[Any]]] = RootModel[
        Annotated[
            str,
            StringConstraints(
                strip_whitespace=True, to_lower=True, min_length=64, max_length=64
            ),
        ]
    ]


AbsolutePathT = TypeVar(
    "AbsolutePathT", bound=Union[HttpUrl, AbsoluteDirectory, AbsoluteFilePath]
)


class RelativePathBase(RootModel[PurePath], Generic[AbsolutePathT], frozen=True):
    _absolute: AbsolutePathT = PrivateAttr()

    @property
    def path(self) -> PurePath:
        return self.root

    @property
    def absolute(self) -> AbsolutePathT:
        """the absolute path/url (resolved at time of initialization with the root of the ValidationContext)"""
        return self._absolute

    def model_post_init(self, __context: Any) -> None:
        if self.root.is_absolute():
            raise ValueError(f"{self.root} is an absolute path.")

        self._absolute = (  # pyright: ignore[reportAttributeAccessIssue]
            self.get_absolute(validation_context_var.get().root)
        )
        super().model_post_init(__context)

    # @property
    # def __members(self):
    #     return (self.path,)

    # def __eq__(self, __value: object) -> bool:
    #     return type(__value) is type(self) and self.__members == __value.__members

    # def __hash__(self) -> int:
    #     return hash(self.__members)

    def __str__(self) -> str:
        return self.root.as_posix()

    def __repr__(self) -> str:
        return f"RelativePath('{self}')"

    @abstractmethod
    def get_absolute(
        self, root: Union[RootHttpUrl, AbsoluteDirectory, pydantic.AnyUrl]
    ) -> AbsolutePathT: ...

    def _get_absolute_impl(
        self, root: Union[RootHttpUrl, AbsoluteDirectory, pydantic.AnyUrl]
    ) -> Union[Path, HttpUrl]:
        if isinstance(root, Path):
            return (root / self.root).absolute()

        parsed = urlsplit(str(root))
        path = list(parsed.path.strip("/").split("/"))
        rel_path = self.root.as_posix().strip("/")
        if (
            parsed.netloc == "zenodo.org"
            and parsed.path.startswith("/api/records/")
            and parsed.path.endswith("/content")
        ):
            path.insert(-1, rel_path)
        else:
            path.append(rel_path)

        return HttpUrl(
            urlunsplit(
                (
                    parsed.scheme,
                    parsed.netloc,
                    "/".join(path),
                    parsed.query,
                    parsed.fragment,
                )
            )
        )

    @classmethod
    def _validate(cls, value: Union[PurePath, str]):
        if isinstance(value, str) and (
            value.startswith("https://") or value.startswith("http://")
        ):
            raise ValueError(f"{value} looks like a URL, not a relative path")

        return cls(PurePath(value))


class RelativePath(
    RelativePathBase[Union[AbsoluteFilePath, AbsoluteDirectory, HttpUrl]], frozen=True
):
    def get_absolute(
        self, root: "RootHttpUrl | Path | AnyUrl"
    ) -> "AbsoluteFilePath | AbsoluteDirectory | HttpUrl":
        absolute = self._get_absolute_impl(root)
        if (
            isinstance(absolute, Path)
            and validation_context_var.get().perform_io_checks
            and not absolute.exists()
        ):
            raise ValueError(f"{absolute} does not exist")

        return absolute


class RelativeFilePath(RelativePathBase[Union[AbsoluteFilePath, HttpUrl]], frozen=True):
    def get_absolute(
        self, root: "RootHttpUrl | Path | AnyUrl"
    ) -> "AbsoluteFilePath | HttpUrl":
        absolute = self._get_absolute_impl(root)
        if (
            isinstance(absolute, Path)
            and validation_context_var.get().perform_io_checks
            and not absolute.is_file()
        ):
            raise ValueError(f"{absolute} does not point to an existing file")

        return absolute


class RelativeDirectory(
    RelativePathBase[Union[AbsoluteDirectory, HttpUrl]], frozen=True
):
    def get_absolute(
        self, root: "RootHttpUrl | Path | AnyUrl"
    ) -> "AbsoluteDirectory | HttpUrl":
        absolute = self._get_absolute_impl(root)
        if (
            isinstance(absolute, Path)
            and validation_context_var.get().perform_io_checks
            and not absolute.is_dir()
        ):
            raise ValueError(f"{absolute} does not point to an existing directory")

        return absolute


FileSource = Union[FilePath, RelativeFilePath, HttpUrl, pydantic.HttpUrl]
PermissiveFileSource = Union[FileSource, str]

V_suffix = TypeVar("V_suffix", bound=FileSource)
path_or_url_adapter = TypeAdapter(Union[FilePath, DirectoryPath, HttpUrl])


def validate_suffix(
    value: V_suffix, suffix: Union[str, Sequence[str]], case_sensitive: bool
) -> V_suffix:
    """check final suffix"""
    if isinstance(suffix, str):
        suffixes = [suffix]
    else:
        suffixes = suffix

    assert len(suffixes) > 0, "no suffix given"
    assert all(
        suff.startswith(".") for suff in suffixes
    ), "expected suffixes to start with '.'"
    o_value = value
    strict = interprete_file_source(value)

    if isinstance(strict, (HttpUrl, AnyUrl)):
        if strict.path is None or "." not in (path := strict.path):
            actual_suffix = ""
        elif (
            strict.host == "zenodo.org"
            and path.startswith("/api/records/")
            and path.endswith("/content")
        ):
            actual_suffix = "." + path[: -len("/content")].split(".")[-1]
        else:
            actual_suffix = "." + path.split(".")[-1]

    elif isinstance(strict, PurePath):
        actual_suffix = strict.suffixes[-1]
    elif isinstance(strict, RelativeFilePath):
        actual_suffix = strict.path.suffixes[-1]
    else:
        assert_never(strict)

    if (
        case_sensitive
        and actual_suffix not in suffixes
        or not case_sensitive
        and actual_suffix.lower() not in [s.lower() for s in suffixes]
    ):
        if len(suffixes) == 1:
            raise ValueError(f"Expected suffix {suffixes[0]}, but got {actual_suffix}")
        else:
            raise ValueError(
                f"Expected a suffix from {suffixes}, but got {actual_suffix}"
            )

    return o_value


@dataclass(frozen=True, **SLOTS)
class WithSuffix:
    suffix: Union[LiteralString, Tuple[LiteralString, ...]]
    case_sensitive: bool

    def __get_pydantic_core_schema__(
        self, source: Type[Any], handler: GetCoreSchemaHandler
    ):
        if not self.suffix:
            raise ValueError("suffix may not be empty")

        schema = handler(source)
        return core_schema.no_info_after_validator_function(
            self.validate,
            schema,
        )

    def validate(self, value: FileSource) -> FileSource:
        return validate_suffix(value, self.suffix, case_sensitive=self.case_sensitive)


def wo_special_file_name(src: FileSource) -> FileSource:
    if has_valid_bioimageio_yaml_name(src):
        raise ValueError(
            f"'{src}' not allowed here as its filename is reserved to identify"
            + f" '{BIOIMAGEIO_YAML}' (or equivalent) files."
        )

    return src


def _package(value: FileSource, info: SerializationInfo) -> Union[str, Path, FileName]:
    if (packaging_context := packaging_context_var.get()) is None:
        # convert to standard python obj
        # note: pydantic keeps returning Rootmodels (here `HttpUrl`) as-is, but if
        #   this function returns one RootModel, paths are "further serialized" by
        #   returning the 'root' attribute, which is incorrect.
        #   see https://github.com/pydantic/pydantic/issues/8963
        #   TODO: follow up on https://github.com/pydantic/pydantic/issues/8963
        if isinstance(value, Path):
            unpackaged = value
        elif isinstance(value, HttpUrl):
            unpackaged = value
        elif isinstance(value, RelativeFilePath):
            unpackaged = Path(value.path)
        elif isinstance(value, AnyUrl):
            unpackaged = str(value)
        else:
            assert_never(value)

        if info.mode_is_json():
            # convert to json value  # TODO: remove and let pydantic do this?
            if isinstance(unpackaged, Path):
                unpackaged = str(unpackaged)
            elif isinstance(unpackaged, str):
                pass
            else:
                assert_never(unpackaged)
        else:
            warnings.warn(
                "dumping with mode='python' is currently not fully supported for "
                + "fields that are included when packaging; returned objects are "
                + "standard python objects"
            )

        return unpackaged  # return unpackaged file source

    # package the file source:
    # add it to the current package's file sources and return its collision free file name
    if isinstance(value, RelativeFilePath):
        src = value.absolute
    elif isinstance(value, pydantic.AnyUrl):
        src = HttpUrl(str(value))
    elif isinstance(value, HttpUrl):
        src = value
    elif isinstance(value, Path):
        src = value.resolve()
    else:
        assert_never(value)

    fname = extract_file_name(src)
    if fname == packaging_context.bioimageio_yaml_file_name:
        raise ValueError(
            f"Reserved file name '{packaging_context.bioimageio_yaml_file_name}' "
            + "not allowed for a file to be packaged"
        )

    fsrcs = packaging_context.file_sources
    assert not any(
        fname.endswith(special) for special in ALL_BIOIMAGEIO_YAML_NAMES
    ), fname
    if fname in fsrcs and fsrcs[fname] != src:
        for i in range(2, 20):
            fn, *ext = fname.split(".")
            alternative_file_name = ".".join([f"{fn}_{i}", *ext])
            if (
                alternative_file_name not in fsrcs
                or fsrcs[alternative_file_name] == src
            ):
                fname = alternative_file_name
                break
        else:
            raise ValueError(f"Too many file name clashes for {fname}")

    fsrcs[fname] = src
    return fname


include_in_package_serializer = PlainSerializer(_package, when_used="unless-none")
ImportantFileSource = Annotated[
    FileSource,
    AfterValidator(wo_special_file_name),
    include_in_package_serializer,
]
InPackageIfLocalFileSource = Union[
    Annotated[
        Union[FilePath, RelativeFilePath],
        AfterValidator(wo_special_file_name),
        include_in_package_serializer,
    ],
    Union[HttpUrl, pydantic.HttpUrl],
]


def has_valid_bioimageio_yaml_name(src: FileSource) -> bool:
    return is_valid_bioimageio_yaml_name(extract_file_name(src))


def is_valid_bioimageio_yaml_name(file_name: FileName) -> bool:
    for bioimageio_name in ALL_BIOIMAGEIO_YAML_NAMES:
        if file_name == bioimageio_name or file_name.endswith("." + bioimageio_name):
            return True

    return False


def identify_bioimageio_yaml_file_name(file_names: Iterable[FileName]) -> FileName:
    file_names = sorted(file_names)
    for bioimageio_name in ALL_BIOIMAGEIO_YAML_NAMES:
        for file_name in file_names:
            if file_name == bioimageio_name or file_name.endswith(
                "." + bioimageio_name
            ):
                return file_name

    raise ValueError(
        f"No {BIOIMAGEIO_YAML} found in {file_names}. (Looking for '{BIOIMAGEIO_YAML}'"
        + " or or any of the alterntive file names:"
        + f" {ALTERNATIVE_BIOIMAGEIO_YAML_NAMES}, or any file with an extension of"
        + f"  those, e.g. 'anything.{BIOIMAGEIO_YAML}')."
    )


def find_bioimageio_yaml_file_name(path: Path) -> FileName:
    if path.is_file():
        if not is_zipfile(path):
            return path.name

        with ZipFile(path, "r") as f:
            file_names = identify_bioimageio_yaml_file_name(f.namelist())
    else:
        file_names = [p.name for p in path.glob("*")]

    return identify_bioimageio_yaml_file_name(file_names)


def ensure_has_valid_bioimageio_yaml_name(src: FileSource) -> FileSource:
    if not has_valid_bioimageio_yaml_name(src):
        raise ValueError(
            f"'{src}' does not have a valid filename to identify"
            + f" '{BIOIMAGEIO_YAML}' (or equivalent) files."
        )

    return src


def ensure_is_valid_bioimageio_yaml_name(file_name: FileName) -> FileName:
    if not is_valid_bioimageio_yaml_name(file_name):
        raise ValueError(
            f"'{file_name}' is not a valid filename to identify"
            + f" '{BIOIMAGEIO_YAML}' (or equivalent) files."
        )

    return file_name


# types as loaded from YAML 1.2 (with ruyaml)
YamlLeafValue = Union[
    bool, _date, _datetime, int, float, str, None
]  # note: order relevant for deserializing
YamlKey = Union[  # YAML Arrays are cast to tuples if used as key in mappings
    YamlLeafValue, Tuple[YamlLeafValue, ...]  # (nesting is not allowed though)
]
YamlValue = _TypeAliasType(
    "YamlValue",
    Union[YamlLeafValue, List["YamlValue"], Dict[YamlKey, "YamlValue"]],
)
BioimageioYamlContent = Dict[str, YamlValue]
BioimageioYamlSource = Union[PermissiveFileSource, BioimageioYamlContent]


@dataclass
class OpenedBioimageioYaml:
    content: BioimageioYamlContent
    original_root: Union[AbsoluteDirectory, RootHttpUrl]
    original_file_name: FileName


@dataclass
class DownloadedFile:
    path: FilePath
    original_root: Union[AbsoluteDirectory, RootHttpUrl]
    original_file_name: FileName


class HashKwargs(TypedDict):
    sha256: NotRequired[Optional[Sha256]]


StrictFileSource = Union[HttpUrl, FilePath, RelativeFilePath]
_strict_file_source_adapter = TypeAdapter(StrictFileSource)


def interprete_file_source(file_source: PermissiveFileSource) -> StrictFileSource:
    if isinstance(file_source, pydantic.AnyUrl):
        file_source = str(file_source)

    if isinstance(file_source, str):
        return _strict_file_source_adapter.validate_python(file_source)
    else:
        return file_source


def _get_known_hash(hash_kwargs: HashKwargs):
    if "sha256" in hash_kwargs and hash_kwargs["sha256"] is not None:
        return f"sha256:{hash_kwargs['sha256']}"
    else:
        return None


def _get_unique_file_name(url: Union[HttpUrl, pydantic.HttpUrl]):
    """
    Create a unique file name based on the given URL;
    adapted from pooch.utils.unique_file_name
    """
    md5 = hashlib.md5(str(url).encode()).hexdigest()
    fname = extract_file_name(url)
    # Crop the start of the file name to fit 255 characters including the hash
    # and the :
    fname = fname[-(255 - len(md5) - 1) :]
    unique_name = f"{md5}-{fname}"
    return unique_name


def download(
    source: Union[PermissiveFileSource, FileDescr],
    /,
    **kwargs: Unpack[HashKwargs],
) -> DownloadedFile:
    """download `source` URL (or pass local file path)"""
    if isinstance(source, FileDescr):
        return source.download()

    strict_source = interprete_file_source(source)
    if isinstance(strict_source, RelativeFilePath):
        strict_source = strict_source.absolute

    if isinstance(strict_source, PurePath):
        local_source = strict_source
        root: Union[RootHttpUrl, DirectoryPath] = strict_source.parent
    else:
        if strict_source.scheme not in ("http", "https"):
            raise NotImplementedError(strict_source.scheme)

        if settings.CI:
            headers = {"User-Agent": "ci"}
            progressbar = False
        else:
            headers = {}
            progressbar = True

        if settings.user_agent is not None:
            headers["User-Agent"] = settings.user_agent

        downloader = pooch.HTTPDownloader(headers=headers, progressbar=progressbar)
        fname = _get_unique_file_name(strict_source)
        _ls: Any = pooch.retrieve(
            url=str(strict_source),
            known_hash=_get_known_hash(kwargs),
            downloader=downloader,
            fname=fname,
            path=settings.cache_path,
        )
        local_source = Path(_ls).absolute()
        root = strict_source.parent

    return DownloadedFile(
        local_source,
        root,
        extract_file_name(strict_source),
    )


class FileDescr(Node):
    source: ImportantFileSource
    """∈📦 file source"""

    sha256: Optional[Sha256] = None
    """SHA256 checksum of the source file"""

    @model_validator(mode="after")
    def validate_sha256(self) -> Self:
        context = validation_context_var.get()
        if not context.perform_io_checks:
            return self

        local_source = download(self.source, sha256=self.sha256).path
        actual_sha = get_sha256(local_source)
        if self.sha256 is None:
            self.sha256 = actual_sha
        elif self.sha256 != actual_sha:
            raise ValueError(
                f"Sha256 mismatch for {self.source}. Expected {self.sha256}, got "
                + f"{actual_sha}. Update expected `sha256` or point to the matching "
                + "file."
            )

        return self

    def download(self):

        return download(self.source, sha256=self.sha256)


def extract_file_name(
    src: Union[pydantic.HttpUrl, HttpUrl, PurePath, RelativeFilePath],
) -> str:
    if isinstance(src, RelativeFilePath):
        return src.path.name
    elif isinstance(src, PurePath):
        return src.name
    else:
        url = urlparse(str(src))
        if (
            url.scheme == "https"
            and url.hostname == "zenodo.org"
            and url.path.startswith("/api/records/")
            and url.path.endswith("/content")
        ):
            return url.path.split("/")[-2]
        else:
            return url.path.split("/")[-1]


def get_sha256(path: Path) -> Sha256:
    """from https://stackoverflow.com/a/44873382"""
    h = hashlib.sha256()
    b = bytearray(128 * 1024)
    mv = memoryview(b)
    with open(path, "rb", buffering=0) as f:
        for n in iter(lambda: f.readinto(mv), 0):
            h.update(mv[:n])

    sha = h.hexdigest()
    assert len(sha) == 64
    return Sha256(sha)



import ast
import sys
import warnings
from typing import Literal
from urllib.parse import urlparse


def get_ref_url(
    type_: Literal["class", "function"], name: str, github_file_url: str
) -> str:
    """get github url with line range fragment to reference implementation from non-raw github file url

    example:
    >>> get_ref_url(
    ...     "class",
    ...     "Binarize",
    ...     "https://github.com/bioimage-io/core-bioimage-io-python/blob/main/bioimageio/core/prediction_pipeline/_processing.py"
    ... )
    'https://github.com/bioimage-io/core-bioimage-io-python/blob/main/bioimageio/core/prediction_pipeline/_processing.py#L120-L127'
    """
    # hotfix to handle requests not available in pyodide, see
    # https://github.com/bioimage-io/bioimage.io/issues/216#issuecomment-1012422194
    try:
        import requests  # not available in pyodide
    except Exception:
        warnings.warn(
            f"Could not reslove {github_file_url} because requests library is not"
            + " available."
        )
        return "URL NOT RESOLVED"

    assert not urlparse(github_file_url).fragment, "unexpected url fragment"
    look_for = {"class": ast.ClassDef, "function": ast.FunctionDef}[type_]
    raw_github_file_url = github_file_url.replace(
        "github.com", "raw.githubusercontent.com"
    ).replace("/blob/", "/")
    try:
        code = requests.get(raw_github_file_url).text
    except requests.RequestException as e:
        warnings.warn(
            f"Could not resolve {github_file_url} due to {e}. Please check your"
            + " internet connection."
        )
        return "URL NOT RESOLVED"
    tree = ast.parse(code)

    for d in tree.body:
        if isinstance(d, look_for):
            assert hasattr(d, "name")
            if d.name == name:
                assert hasattr(d, "decorator_list")
                start = d.decorator_list[0].lineno if d.decorator_list else d.lineno
                if sys.version_info >= (3, 8):
                    stop = d.end_lineno
                else:
                    stop = d.lineno + 1
                break
    else:
        raise ValueError(f"{type_} {name} not found in {github_file_url}")

    return f"{github_file_url}#L{start}-L{stop}"



from typing import Literal, Optional

from pydantic import Field
from typing_extensions import Annotated

from .._internal.common_nodes import Node
from .._internal.io_basics import AbsoluteFilePath as AbsoluteFilePath
from .._internal.types import ImportantFileSource
from .._internal.url import HttpUrl as HttpUrl
from ..generic.v0_2 import VALID_COVER_IMAGE_EXTENSIONS as VALID_COVER_IMAGE_EXTENSIONS
from ..generic.v0_2 import AttachmentsDescr as AttachmentsDescr
from ..generic.v0_2 import Author as Author
from ..generic.v0_2 import BadgeDescr as BadgeDescr
from ..generic.v0_2 import CiteEntry as CiteEntry
from ..generic.v0_2 import Doi as Doi
from ..generic.v0_2 import GenericDescrBase
from ..generic.v0_2 import LinkedResource as LinkedResource
from ..generic.v0_2 import Maintainer as Maintainer
from ..generic.v0_2 import OrcidId as OrcidId
from ..generic.v0_2 import RelativeFilePath as RelativeFilePath
from ..generic.v0_2 import ResourceId as ResourceId
from ..generic.v0_2 import Uploader as Uploader
from ..generic.v0_2 import Version as Version


class ApplicationId(ResourceId):
    pass


class ApplicationDescr(GenericDescrBase, title="bioimage.io application specification"):
    """Bioimage.io description of an application."""

    type: Literal["application"] = "application"

    id: Optional[ApplicationId] = None
    """Model zoo (bioimage.io) wide, unique identifier (assigned by bioimage.io)"""

    source: Annotated[
        Optional[ImportantFileSource],
        Field(description="URL or path to the source of the application"),
    ] = None
    """The primary source of the application"""


class LinkedApplication(Node):
    """Reference to a bioimage.io application."""

    id: ApplicationId
    """A valid application `id` from the bioimage.io collection."""

    version_number: Optional[int] = None
    """version number (n-th published version, not the semantic version) of linked application"""



import json
import sys
from argparse import ArgumentParser
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, Dict, Literal

from deepdiff import DeepDiff
from pydantic import ConfigDict, TypeAdapter
from typing_extensions import assert_never

import bioimageio.spec


def export_json_schemas_from_type(folder: Path, type_: Any, *, title: str):
    adapter = TypeAdapter(
        type_,
        config=ConfigDict(title=title),
    )
    schema = adapter.json_schema()
    for version in ("v" + "-".join(bioimageio.spec.__version__.split(".")), "latest"):
        write_schema(schema, folder / f"bioimageio_schema_{version}.json")


def write_schema(schema: Dict[str, Any], path: Path):
    with path.open("w") as f:
        json.dump(schema, f, indent=4)

    print(f"written `{path}")


def export_json_schemas(dist: Path):
    assert dist.exists()

    export_json_schemas_from_type(
        dist,
        bioimageio.spec.SpecificResourceDescr,
        title=f"bioimage.io resource description {bioimageio.spec.__version__}",
    )


def parse_args():
    p = ArgumentParser(description="script that generates bioimageio json schemas")
    _ = p.add_argument(
        "command", choices=["check", "generate"], nargs="?", default="generate"
    )
    _ = p.add_argument(
        "--dist", nargs="?", default=str((Path(__file__).parent / "../dist").resolve())
    )
    args = p.parse_args()
    return args


def generate_json_schemas(dist: Path, command: Literal["check", "generate"]):
    dist.mkdir(exist_ok=True)
    if command == "generate":
        export_json_schemas(dist)
    elif command == "check":
        existing_schemas = {
            p.name: p for p in Path(dist).glob("bioimageio_schema_*.json")
        }
        with TemporaryDirectory() as tmp_name:
            dist = Path(tmp_name)
            export_json_schemas(dist)
            generated_schemas = {
                p.name: p for p in dist.glob("bioimageio_schema_*.json")
            }
            missing_generated = set(existing_schemas).difference(set(generated_schemas))
            assert not missing_generated, missing_generated
            generated_in_addition = set(existing_schemas).difference(
                set(generated_schemas)
            )
            assert not generated_in_addition, generated_in_addition
            for name, existing_p in existing_schemas.items():
                with existing_p.open() as f:
                    existing = json.load(f)

                with generated_schemas[name].open() as f:
                    generated = json.load(f)

                diff: Any = DeepDiff(existing, generated)
                assert not diff, diff.pretty()
    else:
        assert_never(command)


if __name__ == "__main__":
    args = parse_args()
    sys.exit(generate_json_schemas(Path(args.dist), args.command))
